"""
Description
-----------
This file contains the URDF converter for transforming
arbitrary URDFs into URDFs that the Drake toolbox can
support.
"""
import coacd
from copy import copy, deepcopy
from doctest import UnexpectedException
from enum import IntEnum
import logging
import os
import trimesh
from typing import List

import numpy as np
from pathlib import Path
from pydrake.all import RigidTransform, RotationMatrix, RollPitchYaw
import shutil
import xml.etree.ElementTree as ET

# Internal Imports
from brom_drake.directories import DEFAULT_BROM_MODELS_DIR
from .config import DrakeReadyURDFConverterConfig, MeshReplacementStrategy
from .file_manager import DrakeReadyURDFConverterFileManager
from .mesh_file_converter import MeshFileConverter
from .util import (
    does_drake_parser_support,
    find_mesh_file_path_in,
    get_mesh_element_in,
    URDF_CONVERSION_LOG_LEVEL_NAME, URDF_CONVERSION_LEVEL,
    tree_contains_transmission_for_joint,
    create_transmission_element_for_joint,
)
from brom_drake.file_manipulation.urdf.shapes.cylinder import CylinderDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_element_creator import URDFElementCreator

class DrakeReadyURDFConverter:
    """
    *Description*
    
    This class is used to convert a URDF file into a Drake-compatible URDF file.
    It makes the following considerations when parsing a URDF file:

    - If a mesh file is not supported by the Drake parser,
      then it will convert it into a .obj file (which is supported).
    - If a joint in the URDF file is not "fixed" (i.e., it is actuated)
      and the parser finds no transmission element for it, then it will add
      a transmission element to the URDF file.

    *Parameters*
        
    original_urdf_filename: str
        The path to the original URDF file.
    
    config: DrakeReadyURDFConverterConfig (optional)
        The configuration for the URDF converter.
        This controls:

        - whether or not to convert unsupported mesh files
        - whether or not to add missing transmissions
        - etc.
    """

    def __init__(
        self,
        original_urdf_filename: str,
        config: DrakeReadyURDFConverterConfig = DrakeReadyURDFConverterConfig(),
    ):
        """
        *Description*
        
        This method initializes the DrakeReadyURDFConverter.

        *Parameters*
        
        original_urdf_filename: str
            The path to the original URDF file.

        config: DrakeReadyURDFConverterConfig (optional)
            The configuration for the URDF converter.
            This controls:
            - whether or not to convert unsupported mesh files
            - whether or not to add missing transmissions
            - etc.
        """
        # Setup
        self.config = config
        self.original_urdf_filename = original_urdf_filename

        # Define the file manager which handles the tricky file paths that
        # we encounter during the conversion process.
        self.file_manager : DrakeReadyURDFConverterFileManager = DrakeReadyURDFConverterFileManager(
            original_urdf_path=Path(self.original_urdf_filename),
            models_directory=Path(DEFAULT_BROM_MODELS_DIR),
            log_file_name=config.log_file_name,
            _output_urdf_path=config.output_urdf_file_path,
        ) 

        # Define conversion data directory in brom
        # (If it already exists, then we will overwrite it with the proper flag set)
        self.file_manager.models_directory = Path(DEFAULT_BROM_MODELS_DIR)
        
        overwrite_old_models = self.config.overwrite_old_models

        if overwrite_old_models:
            self.file_manager.clean_up_models_dir()
        os.makedirs(self.file_manager.models_directory, exist_ok=True)

        # Input Processing
        original_urdf_path = Path(self.original_urdf_filename)
        assert original_urdf_path.exists(), f"URDF file {original_urdf_path} does not exist!"

        if not (".urdf" in original_urdf_path.name):
            raise ValueError("Original URDF file must have a .urdf extension!")

        # Set Up logger
        self.logger = self.create_logger()
        self.log(
            f"Created a new DrakeReadyURDFConverter for file \"{self.original_urdf_filename}\".",
        )

        # Configure coacd, just in case
        coacd.set_log_level(config.coacd_log_level)

        # Keep Track of the Joints In The Diagram
        self.actuated_joint_names = []

    def create_logger(self) -> logging.Logger:
        """
        *Description*
        
        This method configures the `logging` logger for the URDF conversion.
        
        *Returns*

        logging.Logger
            The configured logger.
        """
        # Setup
        log_file_name = self.config.log_file_name
        target_file = self.original_urdf_filename

        # Check if the log level exists in the logging module
        overwrite_old_logs = self.config.overwrite_old_logs
        if overwrite_old_logs and os.path.exists(self.file_manager.output_file_directory() / log_file_name):
            os.remove(self.file_manager.output_file_directory() / log_file_name)

        # Create (or collect) a logger for the given file
        logger = logging.getLogger("DrakeReadyURDFConverter ({target_file})" )

        # Create the file handler
        if not self.file_manager.output_file_directory().exists():
            # Create the parent directory if it does not exist
            self.file_manager.output_file_directory().mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(
            self.file_manager.output_file_directory() / log_file_name,
            mode='w'
        )
        file_handler.setLevel(logging.INFO)

        # Create a formatter and set it for the file handler
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )
        file_handler.setFormatter(formatter)

        # Add the file handler to the logger
        logger.addHandler(file_handler)

        # Avoid duplicate logs
        logger.propagate = False

        # Make sure the logger responds to all messages of level DEBUG and above
        logger.setLevel(logging.DEBUG)  # Set to DEBUG to capture all messages

        return logger

    def convert_collision_element(self, collision_elt: ET.Element) -> List[ET.Element]:
        """
        *Description*
        
        This method will convert the collision element in the URDF file.

        *Parameters*
        
        collision_elt: xml.etree.ElementTree.Element
            The collision element that we want to convert.
        
        *Returns*
        
        List[xml.etree.ElementTree.Element]
            The collision elements that will replace the original collision element.
            Usually, this is a list of length 1, but in some cases (e.g., convex decomposition)
            it may be longer.
        """
        # Setup
        replacement_strategy = self.config.mesh_replacement_strategies.collision_meshes

        supported_strategies = [
            MeshReplacementStrategy.kWithObj,
            MeshReplacementStrategy.kWithMinimalEnclosingCylinder,
            MeshReplacementStrategy.kWithConvexDecomposition,
        ]

        # Input Processing
        if replacement_strategy not in supported_strategies:
            error_message = (
                f"Replacement strategy {replacement_strategy} is not supported for collision elements. "
                + "Supported strategies are: "
                + ", ".join([str(strategy) for strategy in supported_strategies])
            )
            self.log(error_message)
            raise ValueError(error_message)

        # Announce algorithm start
        self.log(
            f"Converting element with tag \"{collision_elt.tag}\" using strategy {replacement_strategy}."
        )
        n_replacements_made = 0
        new_elts = []

        # If the strategy is kWithMinimalEnclosingCylinder,
        # then we can replace the entire collision element and don't need to iterate through the children.
        if replacement_strategy == MeshReplacementStrategy.kWithMinimalEnclosingCylinder:
            n_replacements_made += 1
            new_elts = [self.replace_element_with_enclosing_cylinder(
                deepcopy(collision_elt)
            )]

        elif replacement_strategy == MeshReplacementStrategy.kWithObj:
            # Otherwise, let's slowly build the new element
            # by iterating through the children of the collision element.
            new_elt = ET.Element(collision_elt.tag, collision_elt.attrib)
            for ii, child_ii in enumerate(collision_elt):
                # Check to see if the child is a geometry element
                if child_ii.tag == "geometry":
                    new_geometry_elt = self.convert_geometry_element(child_ii, replacement_strategy)
                    new_elt.append(new_geometry_elt)
                else:
                    # If the child is not a geometry element, then we will just copy it over
                    new_elt.append(deepcopy(child_ii))


            # Add the new element to the list of new elements
            new_elts.append(new_elt)

        elif replacement_strategy == MeshReplacementStrategy.kWithConvexDecomposition:
            # Determine if the collision element has a mesh element within it
            original_mesh_path = find_mesh_file_path_in(collision_elt)
            if original_mesh_path is None:
                self.log(
                    "The collision element does not contain a mesh file. Cannot replace with convex decomposition."
                )
                new_elts.append(deepcopy(collision_elt))
                return new_elts
            
            # If the collision element has a mesh element within it, then we will perform convex decomposition
            original_mesh_elt = get_mesh_element_in(collision_elt)
            scale_as_str = original_mesh_elt.attrib.get("scale", "1 1 1")
            original_mesh_scale = np.array(
                [float(x) for x in scale_as_str.split()]
            )

            # Load the mesh file and perform convex decomposition
            self.log(
                f"Found a mesh file at {original_mesh_path}. Performing convex decomposition."
            )
            # Load the mesh file
            mesh: trimesh.Trimesh = trimesh.load(
                self.file_manager.file_path_in_context_of_urdf_dir(original_mesh_path),
                force="mesh",
            )
            mesh = coacd.Mesh(mesh.vertices, mesh.faces)
            parts = coacd.run_coacd(mesh) # a list of convex hulls

            self.log(
                f"coacd created {len(parts)} convex parts from the mesh file \"{original_mesh_path}\""
            )

            for ii, part_ii in enumerate(parts):
                # print(part_ii)
                # print(type(part_ii))
                # print(len(part_ii))

                # Construct new mesh and choose the path to export it to
                mesh_ii = trimesh.Trimesh(
                    vertices=part_ii[0], faces=part_ii[1]
                )
                mesh_file_name_ii = original_mesh_path.name[:original_mesh_path.name.index(".")]
                mesh_ii_file_relative_path = f"meshes/{mesh_file_name_ii}/{mesh_file_name_ii}_part_{ii}.obj"
                mesh_ii_file_path = self.file_manager.output_file_directory() / mesh_ii_file_relative_path
                
                # Export after checking that the parent directory exists
                os.makedirs(mesh_ii_file_path.parent, exist_ok=True)
                mesh_ii.export(mesh_ii_file_path)

                # Create a new collision element for each part
                collision_element_ii = URDFElementCreator.create_collision_element(
                    name=f"{collision_elt.attrib.get('name', 'collision')}_part_{ii}",
                    mesh_file_path="./" + str(mesh_ii_file_relative_path),
                    mesh_scale=original_mesh_scale,
                )
                new_elts.append(collision_element_ii)

                n_replacements_made += 1

        # Check to see if the collision element has a mesh element within it
        
        # # Iterate through every element of the collision element
        # for ii, child_ii in enumerate(collision_elt):
        #     if child_ii.tag == "geometry": 
        #         # If the child is a geometry element, then we will check if replacement is needed.
        #         geometry_elt = child_ii
                
        #         # Iterate through every element of the geometry element
        #         for jj, child_jj in enumerate(geometry_elt):
        #             if child_jj.tag == "mesh":
        #                 # If the child is a mesh element, then we will check if replacement is needed.
        #                 mesh_elt = child_jj
                        
        #                 # Replace!
        #                 if replacement_strategy == MeshReplacementStrategy.kWithObj:
        #                     new_mesh_elt = self.convert_mesh_element(mesh_elt)
        #                     new_elt[ii][jj] = new_mesh_elt
        #                 elif replacement_strategy == MeshReplacementStrategy.kWithMinimalEnclosingCylinder:
        #                     new_elt = self.replace_element_with_enclosing_cylinder(new_elt)
        #                 # elif replacement_strategy == MeshReplacementStrategy.kWithConvexDecomposition:
                            
        #                 else:
        #                     raise ValueError(
        #                         f"Invalid mesh replacement strategy: {replacement_strategy}"
        #                     )    

        #                 n_replacements_made += 1       

        # Announce the number of replacements made in the logger.
        self.log(
            f"Made {n_replacements_made} replacements in the collision element."
        )

        return new_elts
    
    def convert_geometry_element(
        self,
        geometry_elt: ET.Element,
        replacement_strategy: MeshReplacementStrategy
    ) -> ET.Element:
        """
        *Description*
        
        This method will convert the geometry element in the URDF (i.e., an XML) file
        according to the conversion strategy provided as input.
        
        *Parameters*
        
        geometry_elt: xml.etree.ElementTree.Element
            The geometry element that we want to convert.

        replacement_strategy: MeshReplacementStrategy
            The strategy to use for replacing mesh files.
        
        *Returns*
        
        xml.etree.ElementTree.Element
            The new geometry element.
        """
        # Setup
        new_elt = ET.Element(geometry_elt.tag, geometry_elt.attrib)

        # Algorithm
        self.log(
            f"Converting geometry element with strategy \"{replacement_strategy}\"."
        )

        # Iterate through every element of the geometry element
        for ii, child_ii in enumerate(geometry_elt):
            if child_ii.tag == "mesh":
                # If the child is a mesh element, then we will check if replacement is needed.
                mesh_elt = child_ii

                # Replace!
                new_mesh_elt = self.convert_mesh_element(mesh_elt)
                new_elt.append(new_mesh_elt)

            else:
                # If the child is not a mesh element, then we will just copy it over
                new_elt.append(deepcopy(child_ii))

        return new_elt
    
    def convert_visual_element(self, visual_elt: ET.Element) -> ET.Element:
        """
        *Description*
        
        This method will convert the collision element in the URDF file.

        For now, this is very similar to the collision element handling, but it may
        change in the future.

        *Parameters*
        
        visual_elt: xml.etree.ElementTree.Element
            The visual element that we want to convert.

        *Returns*
        
        xml.etree.ElementTree.Element
            The new visual element.
        """
        # Setup
        new_elt = deepcopy(visual_elt)
        replacement_strategy = self.config.mesh_replacement_strategies.visual_meshes

        # Algorithm
        self.log(
            f"Converting element with tag \"{visual_elt.tag}\" using the \"{replacement_strategy}\" strategy."
        )

        # Search through all children of the visual element
        n_replacements_made = 0
        added_material = False
        for ii, child_ii in enumerate(visual_elt):
            if child_ii.tag == "geometry":
                # If the child is a geometry element, then we will check if replacement is needed.
                geometry_elt = child_ii

                # Iterate through every element of the geometry element
                for jj, child_jj in enumerate(geometry_elt):
                    if child_jj.tag == "mesh":
                        # If the child is a mesh element, then we will check if replacement is needed.
                        mesh_elt = child_jj

                        # Replace!
                        if replacement_strategy == MeshReplacementStrategy.kWithObj:
                            new_mesh_elt = self.convert_mesh_element(mesh_elt)
                            new_elt[ii][jj] = new_mesh_elt
                        # TODO: Implement this
                        # elif replacement_strategy == MeshReplacementStrategy.kDoNotReplace:
                        #     # Copy the associated file into the models directory
                        #     old_mesh_file = mesh_elt.attrib["filename"]

                        #     temp_mfc = MeshFileConverter(
                        #         mesh_file_path=old_mesh_file,
                        #         urdf_dir=original_urdf_dir,
                        #         new_urdf_dir=self.file_manager.output_file_directory(),
                        #     )

                        #     true_old_mesh_file = temp_mfc.true_mesh_file_path()
                        #     new_mesh_full_path = temp_mfc.define_output_path()

                        #     print(f"Copying {true_old_mesh_file} to {new_mesh_full_path}")

                        #     # Copy!
                        #     shutil.copy(true_old_mesh_file, new_mesh_full_path)
                            
                        else:
                            raise ValueError(
                                f"Invalid mesh replacement strategy: {replacement_strategy}"
                            )

                        n_replacements_made += 1

            elif child_ii.tag == "material":
                new_material_elt = self.create_new_material_element(child_ii)

                # replace the material element in new_elt with this one
                new_elt[ii] = new_material_elt
                added_material = True
                if self.config.replace_colors_with is not None:
                    self.log(
                        f"Replaced the color of the material element with RGBA values {self.config.replace_colors_with}."
                    )

        # If material did not exist in the visual element, then we will add it.
        if not added_material:
            # Create a new material element
            new_material_elt = ET.Element("material")
            new_elt.append(new_material_elt)

            transformed_material_elt = self.create_new_material_element(new_material_elt)
            new_elt[-1] = transformed_material_elt
            self.log(
                f"Added a new material element to the visual element with RGBA values {self.config.replace_colors_with}."
            )
            n_replacements_made += 1



            added_material = True

        return new_elt

    def convert_urdf(self) -> Path:
        """
        *Description*
        
        Converts the original URDF file into a Drake-compatible URDF file.

        *Returns*
        
        Path
            The path to the new URDF file.
        """
        # Setup
        original_urdf_path = Path(self.original_urdf_filename)
        original_xml = ET.ElementTree(file=original_urdf_path)

        add_missing_actuators = self.config.add_missing_actuators

        # Use recursive function to convert the tree
        new_trees = self.convert_tree(original_xml)
        assert len(new_trees) == 1, \
            f"Expected to get exactly one new tree after conversion, but got {len(new_trees)}."
        new_tree = new_trees[0]

        # Add transmissions, if needed
        if add_missing_actuators:
            for joint_name in self.actuated_joint_names:
                if tree_contains_transmission_for_joint(new_tree, joint_name):
                    continue
                else:
                    # If transmission doesn't exist in URDF, then add it!
                    transmission_element = create_transmission_element_for_joint(joint_name)
                    ET.indent(transmission_element, space="\t", level=0)
                    new_tree.getroot().append(transmission_element)

        # Output the new tree to a file
        output_urdf_path = self.file_manager.output_urdf_path
        os.makedirs(output_urdf_path.parent, exist_ok=True)

        new_tree.write(output_urdf_path)

        self.log(
            f"Converted URDF file to Drake-compatible URDF file at {output_urdf_path}.\n\n"
        )
        return output_urdf_path

    def convert_mesh_element(self, mesh_elt_in: ET.Element) -> ET.Element:
        """
        *Description*

        This method will convert the mesh element in the URDF file
        into a Drake-compatible mesh element.

        *Parameters*

        mesh_elt_in: xml.etree.ElementTree.Element
            The mesh element that we want to convert.

        *Returns*

        xml.etree.ElementTree.Element
            The new mesh element.
        """
        # Setup
        new_elt = deepcopy(mesh_elt_in)

        # Algorithm
        self.log(
            f"Found a mesh element with filename \"{mesh_elt_in.attrib['filename']}\"."
        )

        # Check the value of the filename
        if not ("filename" in new_elt.attrib):
            raise ValueError(
                f"Found a mesh element that does not contain the \"filename\" attribute.\n"
                + "Brom doesn't know how to handle this!"
            )

        # Filename exists; Let's check to see if it's obj or not
        if does_drake_parser_support(new_elt.attrib["filename"]):
            self.log(
                f"\"{new_elt.attrib['filename']}\" is already supported by the Drake parser. Let's copy it into the right place."
            )
            # Let's try to copy the old file into the new directory
            old_filename = new_elt.attrib["filename"]
            new_filename = self.create_obj_to_replace_mesh_file(old_filename)
            new_elt.set(
                "filename", new_filename,
            )
            self.log(
                f"Copied the Drake-compatible mesh file \"{old_filename}\" to \"{new_filename}\"."
            )
        else:
            # If parser does not support the given filename,
            # then let's try to create one
            old_filename = new_elt.attrib["filename"]
            new_filename = self.create_obj_to_replace_mesh_file(old_filename)
            new_elt.set(
                "filename", new_filename,
            )
            self.log(
                f"Replaced the mesh file \"{old_filename}\" with a Drake-compatible .obj file at \"{new_filename}\"."
            )

        return new_elt

    def convert_tree(
        self,
        current_tree: ET.ElementTree
    ) -> List[ET.ElementTree]:
        """
        *Description*
        
        This recursive method is used to parse each element of the
        xml element tree and replace the components that are not
        supported by the Drake URDF parser.

        *Parameters*
        
        current_tree: xml.etree.ElementTree.ElementTree
            The current tree that we are parsing.
        
        *Returns*
        
        List[xml.etree.ElementTree.ElementTree]
            The new trees that has been converted.
        """
        # Setup
        root = deepcopy(current_tree.getroot())

        # Modify root, if necessary
        new_elements = self.convert_tree_element(root)

        for element_ii in new_elements:
            # Modify Children of the root
            initial_children = list(element_ii)
            for child in initial_children:
                new_children = self.convert_tree(
                    ET.ElementTree(child)
                )
                element_ii.remove(child)

                # Add new children to new root
                for new_child in new_children:
                    # Convert the new child element
                    element_ii.append(new_child.getroot())

        return [
            ET.ElementTree(element_ii) for element_ii in new_elements
        ]


    def convert_tree_element(
        self,
        elt: ET.Element,
    ) -> List[ET.Element]:
        """
        *Description*
        
        This method is used to transform one element in the XML tree into an eleemnt
        that that Drake URDF parser can handle. If necessary, it will create a new
        3d model file and place it into the right directory.

        *Parameters*
        
        elt: xml.etree.ElementTree.Element
            An element in the ElementTree that we would like to convert
            into a Drake-ready element.
        
        *Returns*
        
        List[xml.etree.ElementTree.Element]
            The new elements that has been converted.
        """
        # Setup
        new_elts = []

        # Convert file paths for mesh elements if necessary
        if elt.tag == "collision":
            new_elts = self.convert_collision_element(elt)
        elif elt.tag == "visual":
            new_elts = [self.convert_visual_element(elt)]
        elif elt.tag == "transmission":
            # Ignore transmission elements during the conversion
            new_elts = [deepcopy(elt)]
        elif elt.tag == "joint":
            self.handle_joint_element(elt)
            new_elts = [deepcopy(elt)]
        else:
            # If the element is not a mesh element, then we will just return a copy of it.
            self.log(
                f"Found an element with tag \"{elt.tag}\" that does not require handling. No modification needed..."
            )
            new_elts = [deepcopy(elt)]

        # Otherwise, just return a copy of the previous element.
        return new_elts

    def create_obj_to_replace_mesh_file(
        self,
        mesh_file_name: str,
    ) -> str:
        """
        *Description*
        
        This function will create an .obj file to replace the
        .stl or .dae file that is given in mesh file "mesh_file_name".

        *Parameters*
        
        mesh_file_name: str
            The name of the mesh file that we want to convert.
        
        *Returns*
        
        str
            The name of the new .obj file that has been created.
        """
        # Setup
        original_urdf_dir = Path(self.original_urdf_filename).parent

        # Convert the file
        converter = MeshFileConverter(
            mesh_file_path=mesh_file_name,
            urdf_dir=Path(original_urdf_dir),
            new_urdf_dir=self.file_manager.output_file_directory(),
            logger=self.logger,
        )
        new_mesh_path = converter.convert()

        # Clip off all parts of path that include the exported output
        new_mesh_path = str(new_mesh_path).replace(str(self.file_manager.output_file_directory()), "")

        return "." + str(new_mesh_path)


    def handle_joint_element(self, joint_elt: ET.Element):
        """
        *Description*
        
        This method will handle the joint element in the URDF file.
        Specifically, it will check to see if the joint is actuated
        and (if it is) add it to the list of actuated joints.

        This is useful when attempting to add transmissions to the URDF file
        later on.
        
        *Parameters*
        
        joint_elt: xml.etree.ElementTree.Element
            The joint element that we want to handle.
        
        *Returns*
        
        None
        """
        # Setup

        # Keep track of all FREE joint names

        # If the joint does not have a "type" attribute, then we will assume it is fixed
        # and won't add it.
        if "type" not in joint_elt.attrib:
            return

        # If joint is "fixed", then don't add it to the list of actuated joints
        joint_is_fixed = joint_elt.attrib["type"] == "fixed"
        if joint_is_fixed:
            return

        joint_name = joint_elt.attrib["name"]
        self.actuated_joint_names.append(joint_name)

    def create_new_material_element(self, material_elt: ET.Element):
        """
        *Description*
        
        This method will handle the material element in the URDF file.
        Specifically, it will check to see if the user wants us to assign a color to the 
        material element. If so, then we will assign the color to the material element.

        This is useful when attempting to add transmissions to the URDF file
        later on.
        
        *Parameters*
    
        material_elt: xml.etree.ElementTree.Element
            The material element that we want to handle.
        
        *Returns*
        
        None, but modifies the material element.
        """
        # Setup
        replace_colors_with = self.config.replace_colors_with
        material_elt = deepcopy(material_elt)

        # Input Processing (Abort, if no color replacement is needed)
        if replace_colors_with is None:
            return material_elt
        
        if len(replace_colors_with) != 4:
            raise ValueError(
                "The color replacement list must have exactly 4 elements (RGBA)."
            )
        
        # Algorithm
        # Check to see if the material element has a color element
        color_elt = material_elt.find("color")
        if color_elt is None:
            # If no color element exists, then we will create one
            color_elt = ET.Element("color")
            material_elt.append(color_elt)
        else:
            # If a color element exists, then we will modify it
            material_elt.remove(color_elt)
            color_elt = ET.Element("color")
            material_elt.append(color_elt)

        # Set the color of the material element
        color_elt.set(
            "rgba",
            f"{replace_colors_with[0]} {replace_colors_with[1]} {replace_colors_with[2]} {replace_colors_with[3]}"
        )
        return material_elt

    def log(self, message: str):
        """
        *Description*
        
        Logs a message to the logging.logger.

        *Parameters*
        
        message: str
            The message that we want to log.
        
        *Returns*
        
        None
        """
        self.logger.log(logging.INFO, message)

    def replace_element_with_enclosing_cylinder(self, collision_elt: ET.Element) -> ET.Element:
        """
        *Description*
        
        This method will replace the geometry element with an enclosing cylinder.

        *Parameters*
        
        collision_elt: xml.etree.ElementTree.Element
            The collision element that we want to replace.

        *Returns*
        
        xml.etree.ElementTree.Element
            The new collision element with the enclosing cylinder.
        """
        # Setup
        new_elt = deepcopy(collision_elt)
        original_urdf_dir = Path(self.original_urdf_filename).parent

        # Save the initial pose of the element
        origin_elt = new_elt.find("origin")
        pose0 = RigidTransform(
            R=RollPitchYaw(np.array(
                [float(elt) for elt in origin_elt.attrib["rpy"].split(" ")]
            )).ToRotationMatrix(),
            p=np.array(
                [float(elt) for elt in origin_elt.attrib["xyz"].split(" ")],
            ),
        )

        # Extract the mesh element and the filename that it references
        mesh_elt = new_elt.find("geometry/mesh")
        mesh_file_name = mesh_elt.attrib["filename"]

        # Load the mesh from the mesh element
        converter = MeshFileConverter(
            mesh_file_path=mesh_file_name,
            urdf_dir=Path(original_urdf_dir),
            new_urdf_dir=self.file_manager.output_file_directory(),
            logger=self.logger,
        )
        mesh = trimesh.load_mesh(
            str(converter.urdf_dir / converter.true_mesh_file_path())
        )

        # Create the enclosing cylinder
        cylinder_result = trimesh.bounds.minimum_cylinder(mesh, sample_count=12)
        cylinder_shape = CylinderDefinition(
            radius=cylinder_result['radius'],
            length=cylinder_result['height'],
        )

        # Remove the mesh element and pose element from the collision element, if they exist
        geometry_elt = new_elt.find("geometry")
        if mesh_elt is not None:
            new_elt.remove(geometry_elt)

        origin_elt = new_elt.find("origin")
        if origin_elt is not None:
            new_elt.remove(origin_elt)

        # Add the origin element to the collision element
        transform = cylinder_result['transform']
        transform_as_matrix = np.array(transform)
        transform_as_rt = RigidTransform(
            RotationMatrix(transform_as_matrix[:3, :3]),
            transform_as_matrix[:3, 3],
        )

        pose_as_rt = pose0.multiply(transform_as_rt)

        ET.SubElement(
            new_elt,
            "origin",
            {
                "xyz": f"{pose_as_rt.translation()[0]} {pose_as_rt.translation()[1]} {pose_as_rt.translation()[2]}",
                "rpy": f"{pose_as_rt.rotation().ToRollPitchYaw().roll_angle()} {pose_as_rt.rotation().ToRollPitchYaw().pitch_angle()} {pose_as_rt.rotation().ToRollPitchYaw().yaw_angle()}",
            },
        )

        # Create a new geometry element and add the cylinder element to it
        new_elt.append(ET.Element("geometry"))
        cylinder_shape.add_geometry_to_element(new_elt.find("geometry"))

        return new_elt



from dataclasses import dataclass
import networkx as nx

# Internal Imports
from brom_drake.utils.leaf_systems.network_fsm import (
    NetworkXFSM, FSMOutputDefinition, FSMTransitionCondition, FSMTransitionConditionType
)
from .phases import AttemptGraspWithStaticWristPhase

# Define some internal classes/enums
@dataclass
class Script:
    """
    Description
    -----------
    This class defines the "script" or the sequence of events that is meant to happen in the
    AttemptGrasp production.
    """
    settling_time_on_floor: float = 2.0
    grasp_closing_time: float = 5.0
    post_grasp_settling_time: float = 1.0
    drop_time: float = 10.0

    def add_all_states_to_networkx_graph(self, graph: nx.DiGraph) -> list[int]:
        # Setup

        # First State: Object settled
        graph.add_node(
            AttemptGraspWithStaticWristPhase.kObjectSettlingOnFloor,
            outputs=[
                FSMOutputDefinition("start_floor", False),  # Floor trigger
                FSMOutputDefinition("start_gripper", False),  # Gripper trigger
            ]
        )

        # Third State: Close gripper
        graph.add_node(
            AttemptGraspWithStaticWristPhase.kGripperClosing,
            outputs=[
                FSMOutputDefinition("start_gripper", True),
            ]
        )

        # Fourth State: Drop the floor
        graph.add_node(
            AttemptGraspWithStaticWristPhase.kFloorDrop,
            outputs=[
                FSMOutputDefinition("start_floor", True),
            ]
        )

    def start_time_of_phase(self, phase: AttemptGraspWithStaticWristPhase) -> float:
        if phase == AttemptGraspWithStaticWristPhase.kObjectSettlingOnFloor:
            return 0.0
        elif phase == AttemptGraspWithStaticWristPhase.kGripperClosing:
            return self.settling_time_on_floor
        elif phase == AttemptGraspWithStaticWristPhase.kObjectSettlingInGrasp:
            return self.settling_time_on_floor + self.grasp_closing_time
        elif phase == AttemptGraspWithStaticWristPhase.kFloorDrop:
            return self.settling_time_on_floor + self.grasp_closing_time + self.post_grasp_settling_time
        else:
            raise ValueError(f"Phase {phase} not recognized!")

    def to_fsm(self) -> NetworkXFSM:
        return NetworkXFSM(self.to_networkx_graph())

    def to_networkx_graph(self) -> nx.DiGraph:
        # Setup
        graph = nx.DiGraph()

        # Create NetworkX FSM to trigger the floor trajectory
        self.add_all_states_to_networkx_graph(graph)

        # Create time from init -> Let object settle
        graph.add_edge(
            AttemptGraspWithStaticWristPhase.kObjectSettlingOnFloor,
            AttemptGraspWithStaticWristPhase.kGripperClosing, 
            conditions=[
                FSMTransitionCondition(
                    condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                    condition_value=self.settling_time_on_floor,
                )
            ]
        )

        if self.post_grasp_settling_time > 0.0:
            # Create time from gripper start to end of grasp closing
            graph.add_edge(
                AttemptGraspWithStaticWristPhase.kGripperClosing,
                AttemptGraspWithStaticWristPhase.kObjectSettlingInGrasp,
                conditions=[
                    FSMTransitionCondition(
                        condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                        condition_value=self.grasp_closing_time,
                    )
                ]
            )

            # Create time from end of grasp closing to floor drop start
            graph.add_edge(
                AttemptGraspWithStaticWristPhase.kObjectSettlingInGrasp,
                AttemptGraspWithStaticWristPhase.kFloorDrop,
                conditions=[
                    FSMTransitionCondition(
                        condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                        condition_value=self.post_grasp_settling_time,
                    )
                ]
            )

        else:
            # Create time from end of grasp closing to end of production
            graph.add_edge(
                AttemptGraspWithStaticWristPhase.kGripperClosing,
                AttemptGraspWithStaticWristPhase.kFloorDrop,
                conditions=[
                    FSMTransitionCondition(
                        condition_type=FSMTransitionConditionType.kAfterThisManySeconds,
                        condition_value=self.grasp_closing_time,
                    )
                ]
            )

        return graph

    def total_time(self) -> float:
        return self.settling_time_on_floor + self.grasp_closing_time + self.post_grasp_settling_time + self.drop_time
    


from pydrake.all import (
    Context,
    Diagram,
    MultibodyPlant,
    PenetrationAsPointPair,
    QueryObject,
    SceneGraph,
)
from typing import List

def using_point_pair_penetration(
    plant: MultibodyPlant,
    scene_graph: SceneGraph,
    root_context: Context,
    min_acceptable_distance: float = 1e-3,
    debug_flag: bool = False
) -> bool:
    """
    Description
    -----------
    Check for any collisions using the point pair penetration method.
    Returns True if any collisions are detected, otherwise False.

    Arguments
    ---------
    plant : MultibodyPlant
        The plant to check for collisions. (Seems unnecessary)
    scene_graph : SceneGraph
        The scene graph associated with the plant.
    root_context : Context
        The context of the root diagram.
    min_acceptable_distance : float
        The minimum distance to consider as a collision. Defaults to 1e-2.
    """
    # Setup
    scene_graph_context = scene_graph.GetMyContextFromRoot(root_context)
    query_object: QueryObject = scene_graph.get_query_output_port().Eval(scene_graph_context)

    # Check for collisions
    closest_points: List[PenetrationAsPointPair] = query_object.ComputePointPairPenetration()
    collision_detected = False
    for pair_ii in closest_points:
        if pair_ii.depth > min_acceptable_distance:
            collision_detected = True
            if debug_flag:
                print(f"Collision detected between {pair_ii.id_A} and {pair_ii.id_B} with depth {pair_ii.depth}")

            break

    return collision_detected


from dataclasses import dataclass, field
import numpy as np
from pydrake.all import (
    Context,
    Frame,
    ModelInstanceIndex,
    MultibodyPlant,
    RigidTransform,
)
from typing import List

# Internal Imports
from brom_drake.utils.initial_condition_manager.initial_condition import InitialCondition

@dataclass
class InitialConditionManager:
    _ic_tuples: List[InitialCondition] = field(default_factory=list)

    def add_initial_condition(self, ic_tuple: InitialCondition):
        self._ic_tuples.append(ic_tuple)

    def add_initial_configuration(self, model_instance_index: ModelInstanceIndex, configuration: np.ndarray):
        ic_tuple = InitialCondition(
            model_instance_index=model_instance_index,
            configuration=configuration
        )
        self._ic_tuples.append(ic_tuple)

    def add_initial_pose(self, model_instance_index: ModelInstanceIndex, pose_wrt_parent: RigidTransform):
        ic_tuple = InitialCondition(
            model_instance_index=model_instance_index,
            pose_wrt_parent=pose_wrt_parent
        )
        self._ic_tuples.append(ic_tuple)

    def set_all_initial_conditions(self, plant: MultibodyPlant, diagram_context: Context = None):
        """
        *Description*
        
        Set all initial conditions (both the pose and configuration) 
        for the given plant.
        """
        self.set_all_initial_configurations(plant, diagram_context=diagram_context)
        self.set_all_initial_poses(plant, diagram_context=diagram_context)

    def set_all_initial_configurations(self, plant: MultibodyPlant, diagram_context: Context = None):
        """Set all known initial configurations for the given plant."""
        for ic_tuple in self._ic_tuples:
            ic_tuple.set_initial_configuration(plant, diagram_context=diagram_context)


    def set_all_initial_poses(self, plant: MultibodyPlant, diagram_context: Context = None):
        """Set all known initial poses for the given plant."""
        for ic_tuple in self._ic_tuples:
            ic_tuple.set_initial_pose(plant, diagram_context=diagram_context)


from functools import partial
import logging
import networkx as nx
import numpy as np
from pathlib import Path
from pydrake.all import (
    AbstractValue,
    BasicVector,
    Context,
    DiscreteValues,
    LeafSystem,
    PortDataType,
)
from pydrake.systems.framework import OutputPort
from typing import Callable, Dict, List, Union

# Internal Imports
from brom_drake.utils.leaf_systems.network_fsm.fsm_output_definition import FSMOutputDefinition
from brom_drake.utils.leaf_systems.network_fsm.fsm_transition_condition import (
    FSMTransitionCondition, FSMTransitionConditionType
)
from brom_drake.utils.leaf_systems.network_fsm.fsm_edge_definition import FSMEdgeDefinition
from brom_drake.utils.leaf_systems.network_fsm.errors import (
    EdgeContainsNoConditionsError,
    EdgeContainsNoLabelsError,
    MultipleConnectedComponentsError,
    NumberOfStartNodesError,
    OutputPortNotInitializedError,
)
from brom_drake.utils.leaf_systems.network_fsm.config import NetworkXFSMConfig
from brom_drake.directories import DEFAULT_NETWORKX_FSM_DIR

class NetworkXFSM(LeafSystem):
    def __init__(
        self,
        fsm_graph: nx.DiGraph,
        config: NetworkXFSMConfig = NetworkXFSMConfig(),
    ):
        LeafSystem.__init__(self)

        # Setup
        self.config = config

        # Prepare to Create FSM from Graph
        self.fsm_graph = fsm_graph

        # Configure the logger
        self.logger = self.create_logger()

        # Check if the FSM is valid
        exceptions_during_check = self.supports_this_digraph(self.fsm_graph)
        if exceptions_during_check is not None:
            raise exceptions_during_check

        # Collect all edge definitions from the graph
        self.edge_definitions = self.collect_edge_definitions(self.fsm_graph)

        # Create the Input and Output Ports required by the graph
        # NOTE: We will create more than just these ports.
        self.input_port_dict = {}
        self.derive_input_ports_from_graph()

        self.last_output_value = {}
        self.initialize_last_output_value_map()

        self.output_port_dict = {}
        self.allocation_lambdas = {}
        self.output_functions = {}
        self.derive_output_ports_from_graph()

        # Initialize the system in the initial state
        self.t_start_of_current_state = 0.0
        self.current_state_index = None
        self.initialize_fsm_state()

        # Announce completion of initialization
        self.logger.info("Finished initializing NetworkXFSM.")
        self.logger.info("Output Ports Created:")
        for output_port_name in self.output_port_dict:
            self.logger.info(f"+ {output_port_name}")
            output_port_i: OutputPort = self.output_port_dict[output_port_name]

            self.logger.info(f"  - Type: {output_port_i.get_data_type()}")
            if output_port_i.get_data_type() == PortDataType.kVectorValued:
                self.logger.info(f"  - Size: {output_port_i.size()}")
            else:
                self.logger.info(f"  - Allocated Value: {output_port_i.Allocate()}")

        self.logger.info("Last Output Value Map Initialized:")
        for key in self.last_output_value:
            self.logger.info(f"+ {key}: {self.last_output_value[key]}")

    def advance_state_if_necessary(self, context: Context, debug_flag: bool = False):
        """
        *Description*
        
        Checks the current values of all relevant inputs
        and advances the state if necessary according to the
        conditions established in the FSM graph.
        """
        # Setup
        s_t = context.get_discrete_state(self.current_state_index).GetAtIndex(0)
        input_port_dict = self.input_port_dict
        edge_definitions = self.edge_definitions

        # Collect all edge definitions for the current state
        s_t_edge_definitions = [
            edge_definition
            for edge_definition in edge_definitions
            if edge_definition.src == s_t
        ]

        if debug_flag:
            self.log(f"Current state: {s_t}", level=logging.DEBUG)
            self.log(f"Edge definitions for state {s_t}: {s_t_edge_definitions}", level=logging.DEBUG)

        # Iterate through all edge definitions to find the next state (if it exists)
        edge_conditions_satisfied = [False for _ in range(len(s_t_edge_definitions))]
        for ii, edge_definition in enumerate(s_t_edge_definitions):
            # Check all conditions in the edge definition
            all_conditions_met_for_ii = True
            for jj, condition in enumerate(edge_definition.conditions):
                # Check the condition
                condition_ii_met = self.evaluate_transition_condition(
                    condition, context,
                )

                # If the condition is not met,
                # then mark the edge as not satisfied and break the inner loop.
                if not condition_ii_met:
                    all_conditions_met_for_ii = False
                    break

            # If all conditions are met, then mark the edge as satisfied
            edge_conditions_satisfied[ii] = all_conditions_met_for_ii

        if debug_flag:
            self.log(
                f"Edge conditions satisfied: {edge_conditions_satisfied}",
                level=logging.DEBUG,
            )

        # Check the number of conditions that are satisfied
        num_edges_that_are_triggered = sum(edge_conditions_satisfied)
        if (num_edges_that_are_triggered >= 1) and debug_flag:
            self.log(
                f"Number of conditions satisfied: {num_edges_that_are_triggered} at time {context.get_time()}",
                level=logging.DEBUG,
            )

        if num_edges_that_are_triggered == 0:
            # If no conditions are satisfied, then stay in the same state
            next_state = s_t
        elif num_edges_that_are_triggered == 1:
            # If exactly one condition is satisfied, then transition to the next state
            next_state = s_t_edge_definitions[edge_conditions_satisfied.index(True)].dst

            # And update the following variables:
            # - transition time
            # - previous state
            self.t_start_of_current_state = context.get_time()
            context.SetDiscreteState(
                group_index=self.previous_state_index,
                xd=np.array([s_t])
            )
        else:
            # If more than one condition is satisfied, then raise an error
            raise ValueError(
                f"More than one condition is satisfied for state {s_t} in the FSM graph."
            )
            # TODO(kwesi): make this error more verbose to explain to people how to create
            #              mutually exclusive conditions for transition.
        
        if debug_flag:
            print(f"Transitioning from state {s_t} to state {next_state} at time {context.get_time()}")

        # Update the current state
        # self.current_state = next_state
        context.SetDiscreteState(
            group_index=self.current_state_index,
            xd=np.array([next_state])
        )

    def universal_allocation_lambda(self, port_name: str) -> AbstractValue:
        """
        *Description*
        
        This function returns the allocation lambda for the given port name.
        """
        if port_name not in self.allocation_lambdas:
            raise ValueError(f"Allocation lambda for port {port_name} cannot be created because we have no entry for it in the last_output_value dictionary!")
        
        return AbstractValue.Make(self.last_output_value[port_name])

    def CalcFSMState(self, context: Context, fsm_state: BasicVector):
        """
        *Description*
        
        This function calculates the current state of the FSM.
        It uses the various Edge Definitions to determine the next state,
        based on the current state and the input values.
        """
        # Call advance function
        self.advance_state_if_necessary(context)

        # Get the current state
        s_t = context.get_discrete_state(self.current_state_index)

        # Set the FSM state
        fsm_state.SetFrom(s_t)

    def collect_edge_definitions(
        self,
        fsm_graph: nx.DiGraph,
        debug_flag: bool = False,
    ) -> List[FSMEdgeDefinition]:
        """
        Description
        -----------
        This function collects the edge definitions from the FSM graph.
        It uses the data from the graph to create an internal representation
        (FSMEdgeDefinition) for each edge in the graph.
        """
        # Setup
        edge_definitions = []

        # Announce the beginning of the edge collection
        self.logger.info(
            "Collecting edge definitions from the FSM graph...",
        )

        # Iterate through all edges of the FSM graph
        # and identify:
        #   - The input ports that are required for each edge
        for ii, edge_ii in enumerate(fsm_graph.edges):
            src_ii, dst_ii = edge_ii 
            self.log(f"- Edge: {src_ii} -> {dst_ii}", level=logging.INFO)

            edge_data = fsm_graph.edges[edge_ii]

            # Get the conditions for each edge
            ii_th_conditions = edge_data["conditions"]
            
            # Create the FSMEdgeDefinition object
            edge_definition = FSMEdgeDefinition(
                conditions=ii_th_conditions,
                src=src_ii,
                dst=dst_ii,
            )

            # Add the edge definition to the list
            edge_definitions.append(edge_definition)

        return edge_definitions

    def create_logger(self) -> logging.Logger:
        """
        *Description*
        
        This function configures the logger for the FSM.
        It sets the logger to log messages at the INFO level.
        """
        # Setup
        config = self.config

        # Create logger
        logger = logging.getLogger(config.name + " logger")

        # Create a console handler
        if config.show_logs_in_terminal:
            console_handler = logging.StreamHandler()
            console_handler.setLevel(logging.WARNING)

            # Create a formatter and set it for the handler
            formatter = logging.Formatter('%(asctime)s | %(levelname)s | %(name)s | %(message)s')
            console_handler.setFormatter(formatter)

            # Add the handler to the logger
            logger.addHandler(console_handler)

        # Create a file handler
        if config.log_file_name is not None:
            # Create directory if it doesn't exist
            networkx_fsm_log_dir = Path(DEFAULT_NETWORKX_FSM_DIR)
            networkx_fsm_log_dir.mkdir(parents=True, exist_ok=True)

            # Create a file handler
            file_handler = logging.FileHandler(
                filename=DEFAULT_NETWORKX_FSM_DIR + "/" + config.log_file_name,
                mode='w'
            )
            file_handler.setLevel(logging.INFO) # Set to DEBUG to get REALLY verbose logs

            # Create a formatter and set it for the handler
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            file_handler.setFormatter(formatter)

            # Add the handler to the logger
            logger.addHandler(file_handler)

        # Avoid duplicate logs
        logger.propagate = False

        # Make sure the logger responds to all messages of level DEBUG and above
        logger.setLevel(logging.DEBUG)  # Set to DEBUG to capture all messages

        return logger

    def create_output_port_function(
        self,
        port_name_ii: str,
        create_abstract_port: bool = False
    ) -> Callable[[Context, AbstractValue], None]:
        """
        *Description*
        
        This function creates the output port function for the FSM.
        """
        # Setup

        # Create the output port function
        update_map_ii = self.derive_output_update_map(port_name_ii) # (This describes when to update the value of port_ii)

        if create_abstract_port:
            def dummy_output_function_ii(context: Context, output: AbstractValue):
                self.advance_state_if_necessary(context)
                # Get Current State
                s_t = context.get_discrete_state(self.current_state_index).GetAtIndex(0)

                # Check to see if the state value has changed
                if s_t in update_map_ii:
                    # Update the output value
                    self.last_output_value[port_name_ii] = update_map_ii[s_t]

                self.logger.debug(
                    f"Output value for port {port_name_ii} is {self.last_output_value[port_name_ii]} at time {context.get_time()}",
                    )
                output.SetFrom(AbstractValue.Make(self.last_output_value[port_name_ii]))

            self.output_functions[port_name_ii] = dummy_output_function_ii
        else:
            def dummy_output_function_ii(context: Context, output: BasicVector):
                self.advance_state_if_necessary(context)
                # Get Current State
                s_t = context.get_discrete_state(self.current_state_index).GetAtIndex(0)

                # Check to see if the state value has changed
                if s_t in update_map_ii:
                    # Update the output value
                    self.last_output_value[port_name_ii] = update_map_ii[s_t]
                    
                output.SetFrom(self.last_output_value[port_name_ii])

            self.output_functions[port_name_ii] = dummy_output_function_ii
        

        return self.output_functions[port_name_ii]

    def derive_input_ports_from_graph(self, debug_flag: bool = False):
        """
        *Description*
        
        This function creates the input ports for the FSM.
        """
        # Setup
        fsm_graph = self.fsm_graph

        # Iterate through all edges of the FSM graph
        # and identify:
        #   - The input ports that are required for each edge
        input_port_definitions = []
        for ii, edge_ii in enumerate(fsm_graph.edges):
            edge_data = fsm_graph.edges[edge_ii]
            ii_th_conditions = edge_data["conditions"]

            # Get the input port names
            for jj, condition_jj in enumerate(ii_th_conditions):
                # Log what we've found
                self.logger.info(f"Condition {jj} of edge {ii} has:\n" +
                        f"- Input port \"{condition_jj.input_port_name}\"\n" +
                        f"- Condition value \"{condition_jj.condition_value}\"" +
                        f"- Condition type \"{condition_jj.type}\"",
                        )

                if not condition_jj.requires_input_port():
                    self.log(f"Condition {jj} does not require an input port.", level=logging.INFO)

                    continue

                input_port_name = condition_jj.input_port_name
                if input_port_name not in [port_defn[0] for port_defn in input_port_definitions]:
                    input_port_definitions.append(
                        (input_port_name, condition_jj.condition_value)
                    )

        # After collecting all input port names, create the input ports
        for ii, input_port_definition in enumerate(input_port_definitions):
            # Debug messages
            self.logger.info(
                f"Creating input port \"{input_port_definition[0]}\" with example value \"{input_port_definition[1]}\"",
            )

            if debug_flag:
                print(f"Found definition for port \"{input_port_definition}\" with example value.")

            # Extract values from the input port definition
            port_name_ii, example_value_ii = input_port_definition

            # Check to see if port exists already
            if port_name_ii in self.input_port_dict:
                continue

            # Create the port
            if type(example_value_ii) == np.ndarray:
                # Compute size of the input port
                input_port_size = len(example_value_ii)

                # Create port
                self.input_port_dict[port_name_ii] = self.DeclareVectorInputPort(
                    port_name_ii,
                    input_port_size,
                )
                self.logger.info(
                    f"Created vector input port \"{port_name_ii}\" of size {input_port_size}.",
                )
            elif type(example_value_ii) == bool:
                # Create port
                self.input_port_dict[port_name_ii] = self.DeclareAbstractInputPort(
                    port_name_ii,
                    AbstractValue.Make(example_value_ii),
                )

                # Announce creation
                self.logger.info(
                    f"Created abstract input port \"{port_name_ii}\" of type bool.",
                )
            else:
                # Raise an error
                raise ValueError(
                    f"Input port type \"{type(example_value_ii)}\" not supported by" + \
                    " NetworkXFSM.create_input_ports()."
                )

    def derive_output_ports_from_graph(self, debug_flag: bool = False):
        """
        *Description*
        
        This method uses the fsm_graph to infer what output ports
        should be created for the FSM.

        Note
        ----
        For speed, this assumes that the graph has already been checked
        for validity.
        Also, this method assumes that all necessary output ports can be
        inferred from the initial node.
        """
        # Setup
        fsm_graph = self.fsm_graph

        # Get the initial node
        start_nodes = [
            node
            for node in fsm_graph.nodes
            if len(list(fsm_graph.predecessors(node))) == 0
        ]
        initial_node = start_nodes[0]
        initial_node_data = fsm_graph.nodes[initial_node]

        # Create output ports
        initial_outputs: List[FSMOutputDefinition] = initial_node_data["outputs"]
        for output_index_ii, output_port_definition in enumerate(initial_outputs):
            # Extract values from the output port definition
            port_name_ii = output_port_definition.output_port_name
            output_value_ii = output_port_definition.output_port_value

            # Check to see if port exists already
            if port_name_ii in self.output_port_dict:
                continue

            # Create the port
            update_map_ii = self.derive_output_update_map(port_name_ii) # (This describes when to update the value of port_ii)
            if type(output_value_ii) == np.ndarray:
                # Compute size of the output port
                output_port_size = len(output_value_ii)

                # Create port
                self.output_port_dict[port_name_ii] = self.DeclareVectorOutputPort(
                    port_name_ii,
                    output_port_size,
                    self.create_output_port_function(port_name_ii, create_abstract_port=False),
                )

                # Announce creation
                self.logger.info(
                    f"Created vector output port \"{port_name_ii}\" of size {output_port_size}.",
                )
            elif (type(output_value_ii) == bool) or (type(output_value_ii) == str):
                # Make sure to save the value we wish to allocate/the allocation function
                self.allocation_lambdas[port_name_ii] = lambda: AbstractValue.Make(self.last_output_value[port_name_ii])
                self.output_port_dict[port_name_ii] = self.DeclareAbstractOutputPort(
                    port_name_ii,
                    partial(self.universal_allocation_lambda, port_name=port_name_ii),
                    self.create_output_port_function(port_name_ii, create_abstract_port=True),
                )

                # Announce creation
                self.logger.info(
                    f"Created abstract output port \"{port_name_ii}\" of type {type(output_value_ii)}.",
                )
            elif (type(output_value_ii) == int) or (type(output_value_ii) == float):
                # Create a np.array from the scalar value
                output_value_ii = np.array([output_value_ii])

                # Create vector port for new output
                self.output_port_dict[port_name_ii] = self.DeclareVectorOutputPort(
                    port_name_ii,
                    1,
                    self.create_output_port_function(port_name_ii, create_abstract_port=False),
                )

                # Announce creation
                self.logger.info(
                    f"Created vector output port \"{port_name_ii}\" of size 1.",
                )
            else:
                # Raise an error
                raise ValueError(
                    f"Output port type \"{type(output_value_ii)}\" not supported by" + \
                    " NetworkXFSM.create_output_ports()."
                )

    def derive_output_update_map(
        self,
        port_name_ii: str
    ) -> Dict[int, Union[str, int]]:
        """
        *Description*
        
        This method constructs a small mapping of WHEN to change the value of
        port_name_ii.

        The key to each element in the mapping is a value of the state (i.e.,
        on this state, change the value of port_name_ii to this value) and
        the value is the value that the state should be changed to when we reach there.

        *Parameters*
        
        port_name_ii: str
            The name of the port for which this mapping applies.
        """
        # Setup
        fsm_graph = self.fsm_graph

        # Collect all nodes relevant to this port
        relevant_nodes = []
        for node_index_ii in fsm_graph.nodes:
            node_ii = fsm_graph.nodes[node_index_ii]
            if "outputs" not in node_ii:
                continue # Skip this node if it doesn't have an "outputs" key

            if port_name_ii in [ output_port_definition.output_port_name for output_port_definition in node_ii["outputs"] ]:
                relevant_nodes.append(node_index_ii)

        # Iterate through each node in the relevant list and create the mapping
        update_map = {}
        for node_ii in relevant_nodes:
            # Get the output port definition
            output_port_definitions: List[FSMOutputDefinition] = fsm_graph.nodes[node_ii]["outputs"]
            
            # Find the output port for our target port
            output_value_ii = None
            for output_port_definition in output_port_definitions:
                if output_port_definition.output_port_name != port_name_ii:
                    continue

                # Otherwise, we have the right output port
                output_value_ii = output_port_definition.output_port_value

            # Create the mapping
            update_map[node_ii] = output_value_ii

        return update_map

    def evaluate_transition_condition(self, condition: FSMTransitionCondition, context: Context) -> bool:
        """
        Description
        -----------
        This function evaluates the condition for a given input port value.
        """
        # Setup

        # port_is_abstract = self.input_port_dict[input_port_name].get_data_type() == PortDataType.kAbstractValued
        # if port_is_abstract:
        #     input_port_value = input_port_value.get_value()

        # Evaluate the condition
        if condition.type == FSMTransitionConditionType.kAfterThisManySeconds:
            # Check the amount of time elapsed
            elapsed_time = context.get_time() - self.t_start_of_current_state
            return elapsed_time >= condition.condition_value
        else:
            # Evaluate the comparison
            input_port_name = condition.input_port_name
            input_port_value = self.input_port_dict[input_port_name].Eval(context)

            return condition.evaluate_comparison(input_port_value)

    def initialize_last_output_value_map(self):
        """
        *Description*
        
        This function takes the initial node and then uses
        all of the output definitions for that node to define
        the initial values for the `last_output_value` map.
        """
        # Setup
        fsm_graph = self.fsm_graph

        # Get the initial node
        start_nodes = [
            node
            for node in fsm_graph.nodes
            if len(list(fsm_graph.predecessors(node))) == 0
        ]
        initial_node = start_nodes[0]
        initial_node_data = fsm_graph.nodes[initial_node]

        # Create the last output value map
        for output_port_definition in initial_node_data["outputs"]:
            # Extract values from the output port definition
            port_name_ii = output_port_definition.output_port_name
            output_value_ii = output_port_definition.output_port_value

            # Create the last output value map
            self.last_output_value[port_name_ii] = output_value_ii

        self.logger.info(f"Initialized last output value map:")
        for port_name, output_value in self.last_output_value.items():
            self.logger.info(f"- Port \"{port_name}\": {output_value}, type {type(output_value)}")

    def initialize_fsm_state(self):
        """
        Description
        -----------
        This function initializes the FSM state by:
        - Creating the output port for the fsm state,
        - Setting the initial state of the FSM (as an internal variable)
        """
        # Setup
        fsm_graph = self.fsm_graph

        # Get the initial node
        start_nodes = [
            node
            for node in fsm_graph.nodes
            if len(list(fsm_graph.predecessors(node))) == 0
        ]
        initial_node = start_nodes[0]

        # Declare Discrete State for the FSM State
        self.current_state_index = self.DeclareDiscreteState(1)
        self.previous_state_index = self.DeclareDiscreteState(1)

        def discrete_state_init(context: Context, discrete_values: DiscreteValues):
            """
            Initialization function for:
            - the current state of the fsm
            - the previous state of the fsm
            """
            discrete_values.SetFrom(
                DiscreteValues([
                    BasicVector(np.array([initial_node]))
                    for _ in range(2)
                ])
            )

        self.DeclareInitializationDiscreteUpdateEvent(discrete_state_init)

        # Create the output port for the FSM state
        self.DeclareVectorOutputPort(
            "fsm_state",
            1,
            self.CalcFSMState,
        )

    def log(self, message: str, level: int = logging.INFO):
        """
        Description
        -----------
        This function logs a message to the logger.
        """
        self.logger.log(level, message)

    @staticmethod
    def supports_this_digraph(
        digraph: nx.DiGraph,
        debug_flag: bool = False,
    ) -> Union[None, ValueError]:
        """
        Description
        -----------
        This function checks if the given directed graph is a valid FSM.

        Parameters
        ----------
        digraph : nx.DiGraph
            The directed graph to check.

        Returns
        -------
        recognized_exception: Union[None, ValueError]
            Either an error if the graph is not a valid FSM, or None if the graph is valid.
        """

        # Check if the graph is connected
        connected_components = list(nx.weakly_connected_components(digraph))
        if debug_flag:
            print(f"Found {len(connected_components)} connected components.")
        
        if len(connected_components) > 1:
            return MultipleConnectedComponentsError(connected_components)


        # Check if the graph has a single start node
        start_nodes = [
            node
            for node in digraph.nodes
            if len(list(digraph.predecessors(node))) == 0
        ]
        if debug_flag:
            print(f"Found {len(start_nodes)} start nodes.")

        if len(start_nodes) != 1:
            return NumberOfStartNodesError(start_nodes)

        # Check that all edges have a unique label
        # TODO: List all edges and verify that each edge has at least one label
        for edge in digraph.edges:
            # print(edge)
            # print(digraph.edges[edge])
            if digraph.edges[edge] == {}:
                return EdgeContainsNoLabelsError(edge)
            
            conditions = digraph.edges[edge]["conditions"]
            if len(conditions) == 0:
                return EdgeContainsNoConditionsError(edge)

        # Check that the set of output ports (defined by the labels of the initial node)
        # contains ALL the output ports defined in the FSM
        initial_node = start_nodes[0]
        initial_node_data = digraph.nodes[initial_node]
        output_ports_defined = initial_node_data["outputs"]
        output_ports_defined = [output_defn.output_port_name for output_defn in output_ports_defined]
        if debug_flag:
            print(f"Output ports defined by the initial node: {output_ports_defined}")

        # Check that all output ports are defined in the FSM
        for node in digraph.nodes:
            node_data = digraph.nodes[node]
            if "outputs" not in node_data:
                continue

            output_ports = node_data["outputs"]
            output_ports = [output_defn.output_port_name for output_defn in output_ports]
            for output_port in output_ports:
                if output_port not in output_ports_defined:
                    return OutputPortNotInitializedError(node, output_port, output_ports_defined)

        # All checks passed
        return None


from dataclasses import dataclass
import numpy as np
from typing import Union
import xml.etree.ElementTree as ET

# Internal Import
from .shape_definition import ShapeDefinition, ShapeEnum

@dataclass
class BoxDefinition(ShapeDefinition):
    """
    *Description*

    Defines a box shape as per URDF specifications.

    *Attributes*

    size: Union[float, list, tuple, np.ndarray]
        The size of the box along the x, y, and z axes.
        If this is a float, it is assumed to be a cube with equal sides.
    """
    size: Union[float, list, tuple, np.ndarray]

    def add_geometry_to_element(self, target_element: ET.Element) -> ET.Element:
        """
        *Description*

        Adds the box geometry as a sub-element to the given target element.

        This method is an implementation of the abstract method defined in the
        ShapeDefinition base class.

        *Parameters*

        target_element: xml.etree.ElementTree.Element
            The target XML element to which the box geometry will be added.

        *Returns*

        xml.etree.ElementTree.Element
            The element `target_element` with the box geometry added as a sub-element.
        """
        # Setup

        # Convert the size to a numpy array
        if isinstance(self.size, (list, tuple, np.ndarray)):
            self.size = np.array(self.size)
        elif isinstance(self.size, float):
            self.size = np.array([self.size, self.size, self.size])
        else:
            raise ValueError(f"Size {self.size} (type \"{type(self.size)}\") not supported.")

        # Return the element
        return ET.SubElement(
            target_element,
            "box",
            {"size": f"{self.size[0]} {self.size[1]} {self.size[2]}"},
        )

    @property
    def type(self) -> ShapeEnum:
        """
        *Description*

        Always returns ShapeEnum.kBox to indicate this is a box shape.

        This method is an implementation of the abstract method defined in the
        ShapeDefinition base class.

        *Returns*

        ShapeEnum
            The shape type, which is always ShapeEnum.kBox for this class.
        """
        return ShapeEnum.kBox


from importlib import resources as impresources
from typing import List, Tuple, Union
from dataclasses import dataclass, field
import networkx as nx
import numpy as np
from pydrake.all import (
    AbstractValue,
    AffineSystem,
    Adder,
    ConstantValueSource,
    FixedOffsetFrame,
    Frame,
    GeometryProperties,
    InputPortIndex,
    IllustrationProperties,
    JointActuator,
    ModelInstanceIndex,
    PassThrough,
    PidController,
    PiecewisePolynomial,
    PiecewisePose,
    PrismaticJoint,
    RigidBodyFrame,
    RigidTransform,
    SceneGraphInspector,
    TrajectorySource,
)
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph, MultibodyPlant
from pydrake.systems.framework import DiagramBuilder
from pydrake.systems.primitives import ConstantVectorSource, VectorLogSink

# Internal Imports
from brom_drake.control.grippers import GripperController, GripperTarget
from brom_drake.file_manipulation.urdf.drakeify import drakeify_my_urdf
from brom_drake.file_manipulation.urdf.shapes.box import BoxDefinition
from brom_drake.file_manipulation.urdf.simple_writer import SimpleShapeURDFDefinition
from brom_drake.motion_planning.systems.open_loop_dispensers.pose_trajectory_source import PoseTrajectorySource
from brom_drake.robots import find_base_link_name_in, GripperType
from brom_drake import robots
from brom_drake.motion_planning.systems import OpenLoopPlanDispenser, OpenLoopPosePlanDispenser
from brom_drake.productions.types.debug import BasicGraspingDebuggingProduction
from brom_drake.productions import ProductionID
from brom_drake.productions.roles.role import Role
from brom_drake.systems.abstract_list_selection_system import AbstractListSelectionSystem
from brom_drake.systems.abstract_port_switch_system import AbstractPortSwitch
from brom_drake.systems.pose_composition import PoseCompositionSystem
from brom_drake.utils import (
    Performer, collision_checking, NetworkXFSM, FlexiblePortSwitch, FSMTransitionCondition,
    FSMOutputDefinition, FSMTransitionConditionType, Puppetmaker, PuppetmakerConfiguration, PuppetSignature,
)
from brom_drake.utils.initial_condition_manager import InitialConditionManager
from brom_drake.utils.leaf_systems import RigidTransformToVectorSystem, define_named_vector_selection_system
from brom_drake.utils.leaf_systems.rigid_transform_to_vector_system.configuration import Configuration as RigidTransformToVectorSystemConfiguration
from brom_drake.utils.model_instances import (
    get_name_of_first_body_in_urdf,
    find_number_of_positions_in_welded_model,
)
from brom_drake.utils.triads import AddMultibodyTriad
from .config import Configuration
from .phases import AttemptGraspWithPuppeteerWristPhase
from .script import Script as AttemptGraspWithPuppeteerWristScript

class AttemptGraspWithPuppeteerWrist(BasicGraspingDebuggingProduction):
    """
    *Description*

    A production where the user can provide:
    
    - An object to be grasped
    - A gripper
    - The joint positions for the gripper to be in when performing the grasp
    - A pose trajectory for a given gripper frame (usually at the base of the gripper)

    and test whether or not the gripper can successfully grasp the object using the desired approach
    trajectory.

    This production uses a Puppetmaker to control the wrist of the gripper, allowing for
    more complex wrist motions during the grasping attempt.

    The order of events is controlled by a *script*
    (an object of :py:class:`the corresponding Script type<brom_drake.productions.debug.grasping.attempt_grasp.with_puppeteer_wrist.script.Script>`)
    accessible by accessing the ``config.script`` variable of this object.

    *Parameters*

    path_to_object: str
        The file path to the URDF/SDF of the object to be grasped.

    gripper_choice: GripperType
        An enum value from GripperType that specifies which gripper to use for the grasping attempt.

    grasp_joint_positions: np.ndarray
        An array of floats representing the joint positions the gripper should be in when performing the grasp.

    X_ObjectGripper_trajectory: List[RigidTransform], optional
        The desired trajectory for the Gripper Frame.
        A sequence of poses of the Gripper Frame (usually attached to the base of the gripper)
        with respect to the world frame.
        Default is None.

    meshcat_port_number: int, optional
        The port number for Meshcat visualization.

    config: Configuration, optional
        The configuration for the production.
        See the Configuration dataclass for more details.
        Default is the standard Configuration.

    *Usage*

    The user can create an instance of this production and simulate it to see if the grasp is successful in the following way: ::

        from brom_drake.all import drakeify_my_urdf, GripperType, MeshReplacementStrategy
        from brom_drake.productions import AttemptGraspWithPuppeteerWrist

        # Create erlenmeyer flask urdf
        model_file = ... # Some path to your urdf file

        # Create the following poses (transforms):
        # - The Grasp Pose (i.e. the target pose of the gripper wrist when grasping the object)
        X_ObjectTarget = RigidTransform(
            p=np.array([-0.08, 0.05, 0.2]),
            rpy=RollPitchYaw(0.0, np.pi/2.0, 0.0),
        )

        # - The Pre-Grasp Pose (i.e. the pose of the gripper wrist just before reaching to grasp the object)
        X_WorldPreGrasp = X_ObjectTarget.multiply(
            RigidTransform(
                p=np.array([0.0, 0.0, -0.2]),
                rpy=RollPitchYaw(0.0, 0.0, 0.0),
            )
        )

        # Create the production
        production = AttemptGraspWithPuppeteerWrist(
            path_to_object=model_file,
            gripper_choice=GripperType.Robotiq_2f_85,
            grasp_joint_positions=np.array([0.7]),
            X_ObjectGripper_trajectory=[X_WorldPreGrasp, X_ObjectTarget],
            meshcat_port_number=7001, # Use None for CI
        )

        # Build with watcher (so we can view the simulation's data in `brom/watcher/plots` which is helpful for debugging)
        diagram, diagram_context = production.add_cast_and_build()
        script = production.config.script

        # Set up simulation
        simulator = Simulator(diagram, diagram_context)
        simulator.set_target_realtime_rate(1.0)
        simulator.set_publish_every_time_step(False)
        simulator.Initialize()
        simulator.AdvanceTo(script.total_time())

    There are also several examples in the `examples/productions/debug/grasping/attempt_grasp_with_puppeteer_wrist` folder.
        
    """
    def __init__(
        self,
        path_to_object: str,
        gripper_choice: GripperType,
        grasp_joint_positions: np.ndarray,
        X_ObjectGripper_trajectory: List[RigidTransform] = None,
        meshcat_port_number: int = None,
        config: Configuration = Configuration(),
    ):
        """
        *Description*
        

        *Parameters*

        path_to_object: str
            The file path to the URDF/SDF of the object to be grasped.

        gripper_choice: GripperType
            An enum value from GripperType that specifies which gripper to use for the grasping attempt.

        grasp_joint_positions: np.ndarray
            An array of floats representing the joint positions the gripper should be in when performing the grasp.
        
        X_ObjectGripper_trajectory: List[RigidTransform]
            The desired trajectory for the Gripper Frame.
            A sequence of poses of the Gripper Frame (usually attached to the base of the gripper)
            with respect to the object being grasped.

        meshcat_port_number: int, optional
            The port number for Meshcat visualization.

        config: Configuration, optional
            The configuration for the production.
            See the Configuration dataclass for more details.
            Default is the standard Configuration.
        """
        # Use the enum to choose the gripper URDF from a set of supported grippers
        if gripper_choice == GripperType.Robotiq_2f_85:
            path_to_gripper = str(
                impresources.files(robots) / "models/robotiq/2f_85_gripper/urdf/robotiq_2f_85.urdf"
            )
        else:
            raise ValueError(f"Gripper type {gripper_choice} not supported for this scene! Create an issue on GitHub if you want it to be!")

        # Handle config
        config.base.meshcat_port_number = meshcat_port_number

        # Call the parent constructor
        super().__init__(
            path_to_object=path_to_object,
            path_to_gripper=path_to_gripper,
            X_ObjectGrasp=X_ObjectGripper_trajectory[-1],
            meshcat_port_number=config.base.meshcat_port_number,
            time_step=config.base.time_step,
            target_body_on_gripper=config.target_body_on_gripper,
            gripper_color=config.gripper_color,
            show_gripper_base_frame=config.show_gripper_base_frame,
            show_collision_geometries=config.show_collision_geometries,
        )

        # This will define the following fields for the production:
        # - path_to_object
        # - path_to_gripper
        # - X_ObjectTarget
        # - meshcat_port_number
        # - time_step
        # - target_body_on_gripper
        # - gripper_color
        # - show_gripper_base_frame
        # - show_collision_geometries
        # - plant
        # - scene_graph

        # TODO(kwesi): Figure out a better way to CHECK the initial joint positions
        # Create INITIAL joint position array
        self.initial_gripper_joint_positions = np.zeros(grasp_joint_positions.shape)# initial_gripper_joint_positions

        # Save the trajectory of the gripper frame
        self.X_ObjectGripper_trajectory = X_ObjectGripper_trajectory
        self.add_frames_for_gripper_base_trajectory(self.X_ObjectGripper_trajectory)
        
        self.grasp_joint_positions = grasp_joint_positions
        self.config = config
        
        # Add Name to plantPlant and Scene Graph for easy simulation
        self.plant.set_name("AttemptDynamicGrasp_plant")

        # Create show me system for holding object in place
        self.show_me_system = None

        self.floor_model_index = None
        self.floor_actuator = None
        self.floor_shape = None

        # Create an executive/brain
        self.executive = self.create_executive_system() # The executive system that coordinates ALL the systems

    def add_floor_controller_and_connect(self, floor_mass: float):
        """
        *Description*

        This method will add a controller to the "actuated floor" in this production.
        The controller will be a PID controller that maintains the floor at a constant height
        during the first few phases of the production, but eventually we will tell the floor to drop.

        *Parameters*

        floor_mass: float
            The mass of the floor object.
        """
        # Setup
        plant: MultibodyPlant = self.plant

        # Find the z position of the floor
        z_floor = self.find_floor_z_via_line_search()

        # Connect a PID Controller to the floor actuator
        kp = np.array([[floor_mass * 9.81 * 1.0e0]]) #Idk how I'm picking this number.
        ki = np.zeros(kp.shape) # 0.1 * np.sqrt(kp)
        kd = 4.0 * np.sqrt(kp)
        floor_controller = self.builder.AddSystem(
            PidController(kp=kp, ki=ki, kd=kd),
        )
        floor_controller.set_name("[Floor] PID Controller")

        # Create a feedforward term that we will add to the PID controller's
        # output to keep the floor at a constant height
        # (i.e., the weight of the object)
        feedforward_term = self.builder.AddSystem(
            ConstantVectorSource(np.array([[floor_mass * 9.81]])),  # Weight of the object
        )
        feedforward_term.set_name("[Floor] Feedforward")

        # Summer to combine the PID output and the feedforward term
        summer = self.builder.AddSystem(
            Adder(2, 1),
        )

        self.builder.Connect(
            floor_controller.get_output_port(0),  # PID output
            summer.get_input_port(0),  # summer input
        )

        self.builder.Connect(
            feedforward_term.get_output_port(0),  # Weight of the object
            summer.get_input_port(1),  # Feedforward term
        )

        # Connect the target height to the PID controller
        target_height_source = self.create_floor_trajectory_source(z_floor=z_floor)

        # Connect the source to the floor actuator
        self.builder.Connect(
            summer.get_output_port(0),
            plant.get_actuation_input_port(self.floor_model_index),
        )

        # Connect the target height to the PID controller
        self.builder.Connect(
            target_height_source.get_output_port(0),
            floor_controller.get_input_port_desired_state(),
        )

        # Connect the current height of the floor to the PID controller
        self.builder.Connect(
            plant.get_state_output_port(self.floor_model_index),
            floor_controller.get_input_port_estimated_state(),
        )

    def add_floor_to_plant(
        self,
        floor_mass: float = 100.0,
        plant: MultibodyPlant = None,
    ) -> Tuple[
        float, List[float], ModelInstanceIndex, JointActuator
    ]:
        """
        *Description*
        
        This method will add a "floor" (i.e., a rectangular prism with fixed mass)
        to the simulation (i.e., to the plant). It also adds an actuated joint
        between the world origin and the floor, that allows the floor to move up and down.

        *Parameters*

        floor_mass: float, optional
            The mass of the floor object. Default is 100.0 kg.

        plant: MultibodyPlant, optional
            The plant to which the floor will be added.
            If None, the production's plant will be used.
            Default is None.

        *Returns*

        floor_mass: float
            The mass of the floor object.

        floor_shape: List[float]
            The shape of the floor object as [length, width, height].

        floor_model_index: ModelInstanceIndex
            The model instance index of the floor in the plant.

        floor_actuator: JointActuator
            The joint actuator that controls the floor's vertical movement.
        """
        # Setup
        if plant is None:
            plant: MultibodyPlant = self.plant

        # Create a box urdf for the floor
        floor_shape = [10.0, 10.0, 0.1] # Length, Width, Height
        floor_geometry_defn = BoxDefinition(size=floor_shape)

        # Create a urdf for the floor
        floor_urdf_defn = SimpleShapeURDFDefinition(
            name="floor",
            shape=floor_geometry_defn,
            mass=floor_mass,
        ) 
        floor_urdf = floor_urdf_defn.write_to_file()

        # Add the floor to the plant
        floor_model_idcs = Parser(plant=plant).AddModels(floor_urdf)
        floor_model_index = floor_model_idcs[0]

        # Create joint + actuator for the floor
        floor_joint = plant.AddJoint(
            PrismaticJoint(
                name="floor_joint",
                frame_on_parent=plant.world_frame(),
                frame_on_child=plant.GetFrameByName(get_name_of_first_body_in_urdf(floor_urdf)),
                axis=[0, 0, 1],
            )
        )

        floor_actuator = plant.AddJointActuator(
            "floor_elevation_actuator",
            floor_joint,
        )

        return floor_mass, floor_shape, floor_model_index, floor_actuator

    def add_frames_for_gripper_base_trajectory(
        self,
        pose_trajectory: List[RigidTransform]
    ):
        """
        *Description*
        
        This method will add visualizations for the frames
        that are the waypoints in the given gripper base trajectory.

        *Parameters*

        pose_trajectory: List[RigidTransform]
            The desired trajectory for the Gripper Frame.
            A sequence of poses of the Gripper Frame (usually attached to the base of the gripper)
            with respect to the object being grasped.
            TODO(Kwesi): Change the name of this parameter to include frame information.
        """
        # Create frames for each pose in the trajectory
        gripper_base_frames = []
        for i, pose_i in enumerate(pose_trajectory):
            frame_i = FixedOffsetFrame(
                name = "Gripper Base Frame Traj Pt " + str(i),
                P= self.plant.world_frame(), # Parent Frame
                X_PF=pose_i,
            )
            frame_i_added = self.plant.AddFrame(frame_i)
            gripper_base_frames.append(frame_i_added)

        for i, gripper_base_frame_i in enumerate(gripper_base_frames):
            # Add a triad for each frame
            AddMultibodyTriad(
                frame=gripper_base_frame_i,
                scene_graph=self.scene_graph,
                length=0.05,
                radius=0.005,
                opacity=1.0,
            )

        return gripper_base_frames

    def add_gripper_controllers_and_connect(
        self,
        gripper_puppet_input: RigidTransformToVectorSystem,
        gripper_joint_input: PassThrough,
    ):
        # Setup
        builder = self.builder
        script = self.config.script
        plant: MultibodyPlant = self.plant

        gripper_base_pose_source = self.add_gripper_sources_and_connect(builder, plant, script)

        # Combine gripper pose source (X_ManipulandGripper) with the current pose of the
        # manipuland in the world (X_WorldManipuland) and send that signal to the puppet (X_WorldGripperDesired)
        builder.Connect(
            gripper_base_pose_source.get_output_port(),
            gripper_puppet_input.get_input_port(),
        )

        debug_sys2 = builder.AddNamedSystem(
            name="[AttemptGrasp] Gripper Puppet Input Debug",
            system=RigidTransformToVectorSystem(
                RigidTransformToVectorSystemConfiguration(
                    name=f"debug_attempt_grasp_gripper_puppet_input",
                    output_format="vector_xyz_euler(rpy)",
                )
            ),
        )
        builder.Connect(
            gripper_base_pose_source.get_output_port(),
            debug_sys2.get_input_port(),
        )

        # Connect controller for all of the joints
        self.add_gripper_joint_controllers_and_connect(gripper_joint_input)

    def add_gripper_sources_and_connect(
        self,
        builder: DiagramBuilder,
        plant: MultibodyPlant,
        script: AttemptGraspWithPuppeteerWristScript,
    ) -> Tuple[AbstractPortSwitch]:
        # Setup
        executive: NetworkXFSM = self.executive

        # Create a trajectory source for the gripper base (X_ManipulandGripper)
        pose_ManipulandGripper_trajectory = script.build_gripper_approach_trajectory_from_waypoints(self.X_ObjectGripper_trajectory)
        gripper_pose_source = builder.AddNamedSystem(
            name="Source for Gripper Pose in Object Frame",
            system=PoseTrajectorySource(pose_ManipulandGripper_trajectory),
        )

        # Create a system that will convert the reference trajectory
        # from the object frame to the world frame
        object_to_world_converter = builder.AddNamedSystem(
            name="Object to World Frame Converter for Gripper Pose",
            system=PoseCompositionSystem(),
        )

        # - Connect the pose of the object in the world to the FIRST input port
        manipuland_body_indices = plant.GetBodyIndices(self.manipuland_index)
        assert len(manipuland_body_indices) == 1, \
            f"This production currently supports manipulands containing one body; please create an issue if you want more support for objects with multiple parts!"

        pose_selector = builder.AddNamedSystem(
            system=AbstractListSelectionSystem(
                index=int(manipuland_body_indices[0]), 
                output_type=RigidTransform
            ),
            name="Selector for Manipuland Pose in World Frame",
        )
        builder.Connect(
            plant.get_body_poses_output_port(),
            pose_selector.get_input_port(),
        )

        # - Connect the selected pose to the FIRST input port of the converter
        builder.Connect(
            pose_selector.get_output_port(),
            object_to_world_converter.get_pose_AB_port(),
        )

        # - Connect the gripper pose source to the SECOND input port of the converter
        builder.Connect(
            gripper_pose_source.get_output_port(),
            object_to_world_converter.get_pose_BC_port(),
        )

        # Create a system that will "remember" the last
        # desired pose of the gripper in world when we switch away from it
        memory_system = builder.AddNamedSystem(
            name="[AttemptGrasp] Gripper Desired Pose Memory",
            system = ConstantValueSource(
                AbstractValue.Make(self.X_ObjectGripper_trajectory[-1])
            ),
        )

        # Create a flexible port switch to choose between
        # the trajectory source and the memory system
        gripper_pose_port_switch: AbstractPortSwitch = builder.AddNamedSystem(
            name="[AttemptGrasp] Gripper Desired Pose Port Switch",
            system=AbstractPortSwitch(selector_type_in=str, input_type=RigidTransform),
        )

        gripper_pose_port_switch.DeclareInputPort(name="trajectory1")
        gripper_pose_port_switch.DeclareInputPort(name="memory2")

        # - Connect the output of the converter to input 0 of the port switch
        builder.Connect(
            object_to_world_converter.get_output_port(),
            gripper_pose_port_switch.GetInputPort("trajectory1"),
        )

        # Connect the output of the memory system to input 1 of the port switch
        builder.Connect(
            memory_system.get_output_port(),
            gripper_pose_port_switch.GetInputPort("memory2"),
        )

        # # Connect the output of the port switch to the input of the memory system
        # builder.Connect(
        #     gripper_pose_port_switch.get_output_port(),
        #     memory_system.get_input_port(),
        # )

        # - The switch is controlled by the executive
        builder.Connect(
            executive.GetOutputPort("switch_to_memory"),
            gripper_pose_port_switch.get_port_selector_input_port(),
        )

        return gripper_pose_port_switch



    def add_gripper_joint_controllers_and_connect(
        self,
        gripper_joint_input: PassThrough
    ):
        # Setup
        builder: DiagramBuilder = self.builder
        plant: MultibodyPlant = self.plant

        # Create a trajectory source for the gripper joints
        gripper_trajectory_source = self.add_gripper_joints_trajectory_system()

        # Create a simple piece of logic to turn on/off the gripper (i.e., to close it)
        gripper_controller = builder.AddSystem(
            GripperController(
                GripperType.Robotiq_2f_85,
                Kp=1e2 * np.eye(2),
            ),
        )
        gripper_controller.set_name("[Gripper] Gripper Controller")

        # Connect the gripper controller to the gripper
        builder.Connect(
            gripper_controller.GetOutputPort("applied_gripper_torque"),
            gripper_joint_input.get_input_port(),
        )

        # Connect plan to the controller
        builder.Connect(
            gripper_trajectory_source.get_output_port(),
            gripper_controller.GetInputPort("gripper_target"),
        )

        # Create a signal that tells the controller to use position control
        gripper_position_control_source = builder.AddSystem(
            ConstantValueSource(
                AbstractValue.Make(GripperTarget.kPosition),
            )
        )
        gripper_position_control_source.set_name("[Gripper] Target Type Signal")
        builder.Connect(
            gripper_position_control_source.get_output_port(),
            gripper_controller.GetInputPort("gripper_target_type"),
        )

        # Create a system that selects ONLY the elements of the gripper state
        # that are relevant to the gripper controller (i.e., the gripper joint
        # positions and velocities, but none of the puppeteer states)
        states_to_pass_to_gripper_controller = [
            state_name 
            for state_name in plant.GetStateNames(self.gripper_model_index)
            if "puppetmaker" not in state_name
        ]

        state_compressor = builder.AddNamedSystem(
            name=f"[AttemptGrasp] Plant State To Gripper State Converter",
            system=define_named_vector_selection_system(
                all_element_names=plant.GetStateNames(self.gripper_model_index),
                sequence_of_names_for_output=states_to_pass_to_gripper_controller,
            )
        )

        builder.Connect(
            plant.get_state_output_port(self.gripper_model_index),
            state_compressor.get_input_port(),
        )

        builder.Connect(
            state_compressor.get_output_port(),
            gripper_controller.GetInputPort("gripper_state"),
        )

    def add_gripper_joints_trajectory_system(self) -> TrajectorySource:
        """
        *Description*

        This method creates a LeafSystem that outputs a PiecewisePolynomial
        trajectory for the gripper joints to follow during the grasping attempt.

        *Returns*

        gripper_trajectory_source: TrajectorySource
            A TrajectorySource system that outputs the desired gripper joint trajectory.
        """
        # Setup
        builder: DiagramBuilder = self.builder
        script: AttemptGraspWithPuppeteerWristScript = self.config.script

        # Collect gripper initial and final positions
        p_gripper0 = self.initial_gripper_joint_positions
        p_gripper_final = self.grasp_joint_positions

        # Find times for each phase
        t_gripper_close_start = script.start_time_of_phase(
            phase=AttemptGraspWithPuppeteerWristPhase.kGripperClosing
        )
        t_object_settling_in_grasp_start = script.start_time_of_phase(
            phase=AttemptGraspWithPuppeteerWristPhase.kObjectSettlingInGrasp
        )

        # Create a piecewise trajectory for the gripper to follow
        joint_position_samples = []
        joint_position_samples.append(p_gripper0.reshape(-1, 1).tolist()) # This wild compression is needed because of bugs in Drake. :'(
        joint_position_samples.append(p_gripper0.reshape(-1, 1).tolist())
        joint_position_samples.append(p_gripper_final.reshape(-1, 1).tolist())
        gripper_trajectory = PiecewisePolynomial.FirstOrderHold(
            samples=joint_position_samples,
            breaks=[0.0, t_gripper_close_start, t_object_settling_in_grasp_start],
        )

        # Add a trajectory source object to the builder
        gripper_trajectory_source = builder.AddSystem(
            TrajectorySource(gripper_trajectory),
        )
        gripper_trajectory_source.set_name("[Gripper] Trajectory Source")
        return gripper_trajectory_source

    def add_puppeteer_for_gripper(self) -> Tuple[Puppetmaker, PuppetSignature]:
        """
        *Description*

        This method will make the gripper model a "puppet" and return
        the puppetmaker and puppet signature for the gripper.

        By making the gripper a puppet, we can then control its wrist
        using the Puppetmaker's controller.

        *Returns*

        puppetmaker0: Puppetmaker
            The Puppetmaker system that CREATES the gripper puppet.

        puppet_signature0: PuppetSignature
            The PuppetSignature that describes the puppet strings
            for the gripper puppet (i.e., the actuators on the puppet).
        """

        # Setup
        plant : MultibodyPlant = self.plant

        # Create a puppet maker
        puppetmaker0 = Puppetmaker(
            plant=plant,
            config=PuppetmakerConfiguration(
                frame_on_parent=plant.world_frame(),
                name="attempt_grasp_puppetmaker",
                sphere_radius=0.02,
                sphere_color=np.array([0.0,1.0,0.0,0.8]), # Change the color of the puppet links to green
            )
        )

        # Add puppet strings for the gripper
        puppet_signature0 = puppetmaker0.add_strings_for(self.gripper_model_index)

        return puppetmaker0, puppet_signature0

    def add_supporting_cast(self):
        """
        *Description*
        
        This method adds the supporting cast to the production.
        This creates the plant (which contains the object being grasped,
        the gripper, the floor), the controllers for the gripper and floor,
        and the puppetmaker for the gripper wrist.

        This method also is required in order to implement the BaseProduction class.
        """
        # Setup

        # Add the object to the builder
        self.add_manipuland_to_plant()

        # Add the gripper to the builder
        self.add_gripper_to_plant()
        maker0, self.puppet_signature = self.add_puppeteer_for_gripper()

        # Add the floor
        floor_mass, self.floor_shape, self.floor_model_index, self.floor_actuator = self.add_floor_to_plant()

        # Connect the plant to the meshcat, if requested
        if self.meshcat_port_number is not None:
            self.connect_to_meshcat()

        # Finalize the plant
        self.plant.Finalize()
        puppet_pose_input, replacement_gripper_actuator_inputs = maker0.add_puppet_controller_for(
            self.puppet_signature,
            self.builder,
            Kp=np.array([1e2, 1e2, 1e2, 1e1, 1e1, 6e0])
        )

        # Add controllers for gripper AND floor
        self.add_gripper_controllers_and_connect(puppet_pose_input, replacement_gripper_actuator_inputs)
        self.add_floor_controller_and_connect(floor_mass)

        # Create defaults for plant
        self.set_initial_conditions()
        # self.add_initial_conditions_to_plant()    

    def connect_executive_to_gripper_puppet_controller(self, trajectory_dispenser: OpenLoopPosePlanDispenser):
        """
        *Description*

        This method connects the executive system to the trajectory dispenser
        (i.e., provides the trigger signal to start the reference trajectory
        for the gripper wrist).
        """
        # Setup
        builder: DiagramBuilder = self.builder
        executive: NetworkXFSM = self.executive

        # Connect the executive to the trajectory dispenser
        builder.Connect(
            executive.GetOutputPort("enable_gripper_approach"),
            trajectory_dispenser.GetInputPort("plan_ready"),  # Trigger input
        )

    def create_executive_system(self) -> NetworkXFSM:
        """
        *Description*
        
        This method creates the executive system for the production.
        The executive system is responsible for coordinating the various
        phases of the production.
        """
        # Setup
        builder: DiagramBuilder = self.builder
        script: AttemptGraspWithPuppeteerWristScript = self.config.script

        # Create the NetworkXFSM from the graph
        executive = builder.AddSystem(script.to_fsm())

        return executive

    def create_floor_trajectory_source(
        self,
        z_floor: float
    ) -> Tuple[FlexiblePortSwitch]:
        """
        *Description*

        This method creates a LeafSystem which outputs a reference trajectory
        for the floor height during the production.

        *Parameters*

        z_floor: float
            The initial height of the floor.

        *Returns*

        floor_target_height_source: FlexiblePortSwitch
            A LeafSystem which outputs a reference trajectory for the floor height during the production.
            It is a "switch" system that can switch between different input ports
            based on the executive's commands.
        """
        # Setup
        builder: DiagramBuilder = self.builder

        #Create a proper trajectory source for the floor so that it slowly falls away
        floor_target_height_source = builder.AddSystem(
            OpenLoopPlanDispenser(2, speed=0.25),
        )

        # Create the trajectory that will be run
        simple_plan = np.array([
            [z_floor, 0.0],
            [z_floor - 10.0, 0.0]
        ])
        floor_trajectory_source = builder.AddSystem(
            ConstantValueSource(AbstractValue.Make(simple_plan))
        )
        
        builder.Connect(
            floor_trajectory_source.get_output_port(),
            floor_target_height_source.GetInputPort("plan"),
        )

        # Connect timer to the plan dispenser
        builder.Connect(
            self.executive.GetOutputPort("start_floor"),
            floor_target_height_source.GetInputPort("plan_ready"),  # Trigger input
        )

        return floor_target_height_source

    def find_floor_z_via_bounding_box(
        self,
        debug_flag: bool = True,
    ) -> float:
        """
        *Description*
        
        This method will find the transform needed to put the floor at so that
        the object sits "just" on top of it.

        WARNING: This method does not work and I don't know why. (i.e., sometimes
        the floor is in collision with the object)

        *Parameters*

        debug_flag: bool, optional
            If True, will print debug information during the computation. Default is True.

        *Returns*

        floor_z_translation: float
            The z translation for the floor such that the object sits just on top of it.
        """
        # Setup
        shadow_builder = DiagramBuilder()

        # Create shadow plant + scene graph
        shadow_plant, shadow_scene_graph = AddMultibodyPlantSceneGraph(
            shadow_builder,
            time_step=self.time_step,
        )
        shadow_plant.set_name("DemonstrateStaticGrasp_shadow_plant")
        shadow_scene_graph.set_name("DemonstrateStaticGrasp_shadow_scene_graph")

        # Add the object to the shadow plant
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_object)
        assert len(model_idcs) == 1, f"Only one model should be added; received {len(model_idcs)}"
        model_idx = model_idcs[0]
        model_name = shadow_plant.GetModelInstanceName(model_idx)

        # Weld the base link to the world frame
        # manipuland_body_idcs = shadow_plant.GetBodyIndices(model_idx)
        # assert len(manipuland_body_idcs) > 0, \
        #     f"Expected at least one body in the manipuland; received {len(manipuland_body_idcs)}"
        # manipuland_body = shadow_plant.get_body(manipuland_body_idcs[0])
        # frame0 = manipuland_body.body_frame()
        # shadow_plant.WeldFrames(
        #     shadow_plant.world_frame(),
        #     frame0,
        # )

        # Finalize the shadow plant
        shadow_plant.Finalize()

        # Search through the geometries in the scene and find the manipuland's
        # geometry. Attempt to compute the bounding box around it.
        shadow_sg_inspector: SceneGraphInspector = shadow_scene_graph.model_inspector()
        all_geometries = shadow_scene_graph.model_inspector().GetAllGeometryIds()
        objects_minimum_z = 1_000.0
        for ii, geometry_ii in enumerate(all_geometries):
            # Announce geometry value:
            if debug_flag:
                print(f"Geometry ID {ii}:")
                print(f"- Raw value: {geometry_ii}")

            # Get shape of the geometry
            shape_ii = shadow_sg_inspector.GetShape(geometry_ii)
            if debug_flag:
                print(f"- Shape {ii}:")
                print(f"  + Raw value: {shape_ii}")
                print(f"  + Type: {type(shape_ii)}")

            # Get the convex hull
            ch_ii = shape_ii.GetConvexHull()
            if debug_flag:
                print(f"- Convex Hull {ii}:")
                print(f"  + Raw value: {ch_ii}")

            bb_ii = ch_ii.CalcBoundingBox()
            if debug_flag:
                print(f"- Bounding Box {ii}:")
                print(f"  + Raw value: {bb_ii}")
            lower_left, upper_right = bb_ii

            if lower_left[2] < objects_minimum_z:
                objects_minimum_z = lower_left[2]
                print(f"  + New z_out: {objects_minimum_z}")
        
        # If we didn't find any geometry, raise an error
        if objects_minimum_z == 1_000.0:
            raise ValueError("Unable to find geometry for the manipuland.")
        
        # If we found a geometry, then we need to set the floor height
        # using the known shape of the table

        return objects_minimum_z - self.floor_shape[2]*0.5  # Subtract the height of the floor

    def find_floor_z_via_line_search(
        self,
        debug_flag: bool = False,
    ) -> float:
        """
        *Description*
        
        Use bounding box method to generate an initial guess for the floor height
        and then perform a line search to find the height that actually
        is collision free.

        *Parameters*

        debug_flag: bool, optional
            If True, will print debug information during the computation. Default is False.

        *Returns*

        floor_z_translation: float
            The z translation for the floor such that the object sits just on top of it
        """
        # Setup
        z_floor = self.find_floor_z_via_bounding_box(debug_flag=debug_flag)
        floor_shape = self.floor_shape

        # Check for collisions at the initial guess
        collision_detected = self.floor_z_creates_collisions(
            floor_z_translation=z_floor,
            debug_flag=debug_flag,
        )
        if not collision_detected:
            # If no collision detected, return the initial guess
            return z_floor
        
        # If we have a collision, then we need to perform a line search
        # to find the height that is collision free
        
        # Start with a small step size
        step_size = floor_shape[2] * 0.25  # 25% of the floor height

        # Perform a line search to find the height that is collision free
        candidate_z_floor = z_floor - step_size
        while True:
            # Check for collisions at the current height
            collision_detected = self.floor_z_creates_collisions(
                floor_z_translation=candidate_z_floor,
                debug_flag=debug_flag,
            )
            if not collision_detected:
                # If no collision detected, return the current height
                print(f"Found collision-free height: {candidate_z_floor}")
                return candidate_z_floor
            

            print(f"Collision detected at height: {candidate_z_floor}")
            # If we have a collision, then increase the height
            candidate_z_floor -= step_size

            # If we exceed a certain height, break out of the loop
            if candidate_z_floor < -10.0:
                break        
                
        raise ValueError("Unable to find a collision-free height for the floor.")

    def find_X_WorldGripper(
        self,
        X_ObjectGripper: RigidTransform,
        target_frame_name: str,
        desired_joint_positions: List[float],
    ) -> RigidTransform:
        """
        *Description*
        
        This method will return the transform from the world frame to the gripper frame.

        *Parameters*

        X_ObjectGripper: RigidTransform
            The desired transform from the object frame to the gripper target frame.

        target_frame_name: str
            The name of the target frame on the gripper.
            (i.e., the frame that we define as "Gripper" in the X_ObjectGripper transform).

        desired_joint_positions: List[float]
            The desired joint positions for the gripper.
            TODO(Kwesi): Is this needed? Maybe when we use the fingers as the target frame name.

        *Returns*

        X_WorldGripper: RigidTransform
            The transform from the world frame to the gripper frame.
        """
        # Setup
        desired_joint_positions = np.array(desired_joint_positions)

        # Create shadow plant containing just the gripper
        shadow_plant = MultibodyPlant(self.time_step)
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_gripper)
        shadow_model_idx = model_idcs[0]
        
        # Get the first body on the gripper
        name_of_first_body_in_gripper = get_name_of_first_body_in_urdf(self.path_to_gripper)

        # Weld the base link to the world frame
        shadow_plant.WeldFrames(
            shadow_plant.world_frame(),
            shadow_plant.GetFrameByName(name_of_first_body_in_gripper),
        )

        # Finalize Plant
        shadow_plant.Finalize()

        shadow_builder = DiagramBuilder()
        shadow_builder.AddSystem(shadow_plant)
        diagram = shadow_builder.Build()
        shadow_diagram_context = diagram.CreateDefaultContext()

        # Set the joint positions
        # shadow_plant.SetPositions(
        #     shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
        #     shadow_model_idx,
        #     desired_joint_positions,
        # )

        # Get the transform of the gripper's base in the static world frame
        X_GripperBase_GripperTarget = shadow_plant.EvalBodyPoseInWorld(
            shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
            shadow_plant.GetBodyByName(target_frame_name),
        )

        X_Object_GripperTarget = X_ObjectGripper
        X_WorldObject = RigidTransform.Identity()
        X_Object_GripperBase = X_Object_GripperTarget.multiply(
            X_GripperBase_GripperTarget.inverse(),
        )

        return X_WorldObject.multiply(X_Object_GripperBase)

    def floor_z_creates_collisions(
        self,
        floor_z_translation: float = 0.0,
        debug_flag: bool = False,
    ) -> bool:
        """
        *Description*
        
        This method will check if the given floor z translation
        creates collisions with the object in the production.

        *Parameters*

        floor_z_translation: float, optional
            The z translation of the floor to check for collisions. Default is 0.0.

        debug_flag: bool, optional
            If True, will print debug information during the computation. Default is False.
            TODO(Kwesi): Implement debug information.

        *Returns*

        collision_detected: bool
            True if a collision is detected between the floor and the object, False otherwise.
        """
        # Setup
        shadow_builder = DiagramBuilder()

        # Create shadow plant + scene graph
        shadow_plant, shadow_scene_graph = AddMultibodyPlantSceneGraph(
            shadow_builder,
            time_step=self.time_step,
        )
        shadow_plant.set_name("DemonstrateStaticGrasp_shadow_plant")
        shadow_scene_graph.set_name("DemonstrateStaticGrasp_shadow_scene_graph")

        # Add the object to the shadow plant
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_object)
        assert len(model_idcs) == 1, f"Only one model should be added; received {len(model_idcs)}"
        model_idx = model_idcs[0]
        model_name = shadow_plant.GetModelInstanceName(model_idx)

        # # Add the gripper to the shadow plant
        # temp_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_gripper)
        # assert len(temp_idcs) == 1, f"Only one model should be added; received {len(temp_idcs)}"
        # shadow_gripper_model_index = temp_idcs[0]
        # shadow_gripper_model_name = shadow_plant.GetModelInstanceName(shadow_gripper_model_index)

        # Add the floor to the shadow plant
        _, _, shadow_floor_model_index, _ = self.add_floor_to_plant(plant=shadow_plant)

        # Finalize the shadow plant
        shadow_plant.Finalize()

        # Create diagram and set position of the floor to be floor_z_translation
        shadow_diagram = shadow_builder.Build()
        shadow_diagram_context = shadow_diagram.CreateDefaultContext()

        shadow_plant.SetPositions(
            shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
            shadow_floor_model_index,
            np.array([floor_z_translation]),
        )

        return collision_checking.using_point_pair_penetration(
            shadow_plant, shadow_scene_graph, shadow_diagram_context,
        )

    @property
    def id(self):
        """
        *Description*

        Always returns ProductionID.kAttemptGraspWithPuppeteer.
        
        This property is required to implement the BaseProduction class.
        """
        return ProductionID.kAttemptGraspWithPuppeteer

    def set_initial_conditions(self):
        """
        *Description*
        
        This method will set the default state of the plant to the desired joint positions.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        ic_manager: InitialConditionManager = self.initial_condition_manager

        # Find the z position of the floor
        z_floor = self.find_floor_z_via_line_search()

        # Set the default state of the floor to be at the desired height
        ic_manager.add_initial_configuration(
            model_instance_index=self.floor_model_index,
            configuration=np.array([z_floor]),
        )

        # Set the initial pose of the object
        ic_manager.add_initial_pose(
            model_instance_index=self.manipuland_index,
            pose_wrt_parent=RigidTransform.Identity(),
        )

        # # - Set the initial configuration of some of the puppet joints to match the gripper
        # #   (i.e., so that the puppet and gripper start in the same position)
        # X_WorldObject = X_WorldObject.multiply(self.X_ObjectGripper)
        # for jj in range(3):
        #     puppet_model_jj = self.puppet_signature.all_models[jj]
        #     ic_manager.add_initial_configuration(
        #         model_instance_index=puppet_model_jj,
        #         configuration=np.array([self.X_World_PreGrasp.translation()[jj]]),
        #     )

        # for jj in range(2):
        #     puppet_model_jj = self.puppet_signature.all_models[jj + 3]
        #     ic_manager.add_initial_configuration(
        #         model_instance_index=puppet_model_jj,
        #         configuration=np.array([self.X_World_PreGrasp.rotation().ToRollPitchYaw().vector()[jj]]),
        #     )

        # Set initial conditions for the puppeteer joints
        X_World_PreGrasp = self.X_ObjectGripper_trajectory[0]
        assert len(self.puppet_signature.all_joint_actuators) == 6, \
            f"Expected 6 puppet joint actuators; received {len(self.puppet_signature.all_joint_actuators)}"

        for joint_index, joint_actuator_i in enumerate(self.puppet_signature.all_joint_actuators):
            if joint_index < 3:
                self.initial_condition_manager.add_initial_configuration(
                    model_instance_index=joint_actuator_i.model_instance(),
                    configuration=np.array([X_World_PreGrasp.translation()[joint_index]])
                )
            elif joint_index == 3: # Yaw
                self.initial_condition_manager.add_initial_configuration(
                    model_instance_index=joint_actuator_i.model_instance(),
                    configuration=np.array([X_World_PreGrasp.rotation().ToRollPitchYaw().yaw_angle()])
                )
            elif joint_index == 4: # Pitch
                self.initial_condition_manager.add_initial_configuration(
                    model_instance_index=joint_actuator_i.model_instance(),
                    configuration=np.array([X_World_PreGrasp.rotation().ToRollPitchYaw().pitch_angle()])
                )

        # Set the initial configuration of the gripper
        # - Set the gripper positions (i.e., initial positions of the joints)
        #   (Because of the puppeteer, this has an additional fixed joint at the base)
        n_gripper_positions1 = plant.num_positions(self.gripper_model_index)

        initial_gripper_configuration = np.zeros((n_gripper_positions1,))
        initial_gripper_configuration[0] = X_World_PreGrasp.rotation().ToRollPitchYaw().roll_angle()        
        initial_gripper_configuration[1:] = self.initial_gripper_joint_positions
        
        ic_manager.add_initial_configuration(
            model_instance_index=self.gripper_model_index,
            configuration=initial_gripper_configuration,
        )

        # # - Set the gripper's base pose
        # X_WorldObject = X_WorldObject.multiply(self.X_ObjectGripper)
        # ic_manager.add_initial_pose(
        #     model_instance_index=self.gripper_model_index,
        #     pose_wrt_parent=X_WorldObject,
        # )

        

    @property
    def suggested_roles(self) -> List[Role]:
        """
        *Description*
        
        This property suggests the roles that are relevant for this production.

        No roles are necessary here, so this method returns an empty list.
        
        *Returns*
        suggested_roles: List[Role]
            A list of Role enums that are relevant for this production.
        """
        return []
    
    


"""
gripper_target.py
Description:

    This file defines the GripperTarget enum and associated values.
"""

from enum import IntEnum


class GripperTarget(IntEnum):
    kPosition = 1
    kVelocity = 2


"""
joint_target.py
Description:

    This file defines the JointTarget enum.
    This class is used to control different manipulators and helps us
    define the type of target that we want to control.
"""
from enum import IntEnum

class JointTarget(IntEnum):
    """
    *Description*
    The type of target being given for the control of a robot's joints.

    *Usage*

    .. code-block:: python

        from brom_drake.control.arms.joint_target import JointTarget
        # or
        # from brom_drake.all import JointTarget

        # Selecting the position target type
        joint_target_type = JointTarget.kPosition
    """
    kPosition = 1
    kVelocity = 2
    kTorque = 3


"""
DiagramWatcher.py
Description:

    Creates a DiagramWatcher class that can be used to monitor the state of a Diagram
    and automatically log signals of interest.
"""
import os
from typing import Dict, List, Union
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import logging

from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, DiagramBuilder, LeafSystem, PortDataType
from pydrake.systems.primitives import (
    VectorLogSink, ConstantVectorSource, AffineSystem, LogVectorOutput,
)

# Internal Imports
from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.PortWatcher.port_watcher import PortWatcher
from brom_drake.PortWatcher.port_watcher_options import PortWatcherOptions, PortWatcherPlottingOptions, PortWatcherRawDataOptions
from brom_drake.DiagramWatcher.diagram_watcher_options import DiagramWatcherOptions
from brom_drake.DiagramWatcher.constants import INELIGIBLE_SYSTEM_TYPES
from brom_drake.DiagramWatcher.errors import UnrecognizedTargetError


class DiagramWatcher:
    """
    *Description*
    
    An object that will iterate through all elements of a partially built
    Drake Diagram (via the DiagramBuilder) and add PortWatchers to the specified targets.

    *Parameters*

    subject: DiagramBuilder
        We will search through the subject (a diagram builder)
        to find all the systems that we want to monitor.

    targets: List[DiagramTarget], optional
        The targets that we want to monitor, by default this tries to monitor
        all systems in the diagram. (i.e., when this value is None, we will monitor all systems).

    options: DiagramWatcherOptions, optional
        The options that configure the DiagramWatcher, by default DiagramWatcherOptions()
    """
    def __init__(
        self,
        subject: DiagramBuilder,
        targets: List[DiagramTarget] = None,
        options: DiagramWatcherOptions = DiagramWatcherOptions(),
    ):
        """
        *Description*
        
        Initializes the DiagramWatcher class.

        *Parameters*
        
        subject : DiagramBuilder
            We will search through the subject (a diagram builder)
            to find all the systems that we want to monitor.

        targets : List[DiagramTarget], optional
            The targets that we want to monitor, by default None

        port_watcher_options : PortWatcherOptions, optional
            The options for the PortWatcher, by default PortWatcherOptions()
        """
        # Setup

        # Needs to be populated by the user of this class AFTER the diagram has been built
        self.diagram = None
        self.diagram_context = None

        # Check subject
        if not isinstance(subject, DiagramBuilder):
            raise ValueError("subject must be a DiagramBuilder!")

        # Save the inputs
        self.subject = subject
        self.options = options

        # Create the .brom directory, to store:
        # - activity_summary.log
        # - all plots
        if os.path.exists(self.options.base_directory):
            os.system(f"rm -r {self.options.base_directory}")

        # Create directory to plot in
        os.makedirs(self.options.base_directory, exist_ok=True)
        self.logger = self.create_logger()  # Create an "activity summary" log
                                                # which details what the
                                                # DiagramWatcher is doing.

        # Collect All the Connections and Systems
        self.eligible_systems = self.find_eligible_systems(subject)
        if targets is None:
            targets = [DiagramTarget(system.get_name()) for system in self.eligible_systems]
        else:
            self.check_targets(targets, self.eligible_systems)

        # Log the list of eligible systems
        self.logger.info(f"Found {len(self.eligible_systems)} systems in diagram are eligible for targeting:")
        for idx, system in enumerate(self.eligible_systems):
            self.logger.info(f"{idx}: {system.get_name()}")

        # For Each Target with None ports, we will try to
        # "smartly" create the targets that we want to monitor
        inferred_targets = self.get_smart_targets(subject, targets)
        self.inferred_targets = inferred_targets

        # For each target's port, we will add a logger
        self.port_watchers: Dict[str, Dict[str, PortWatcher]] = {target.name: {} for target in inferred_targets}
        self.logger.info("Adding loggers to the diagram... (via PortWatcher objects)")
        for target in inferred_targets:
            system = subject.GetSubsystemByName(target.name)
            for port_index in target.ports:
                target_port = system.get_output_port(port_index)
                options_for_target_port = options.to_port_watcher_options()

                try:
                    # Configure PortWatcher
                    self.port_watchers[target.name][target_port.get_name()] = PortWatcher(
                        target_port, subject, self.logger,
                        logger_name=f"{target.name}_logger_{port_index}",
                        base_watcher_dir=options.base_directory,
                        options=options_for_target_port,
                    )

                except Exception as e:
                    if self.options.hide_messages.during_port_watcher_connection:
                        continue

                    # If we have an error, we will log it
                    self.logger.debug(
                        f"Failed to add a watcher to port {target_port.get_name()} of system {target.name}: {e}",
                        exc_info=False,
                    )

                if target_port.get_name() in self.port_watchers[target.name]:
                    # Announce that we successfully added logger
                    self.logger.info(f"Added logger to port {target_port.get_name()} of system {target.name}")
                else:
                    # If we did not add the logger, then we will send a small note.
                    self.logger.info(
                        f"Unable to add logger to port {target_port.get_name()} of system {target.name}",
                    )
                    

    def create_new_port_watcher_options(
        self,
        options: PortWatcherOptions,
        plot_dir: str = None,
        raw_data_dir: str = None,
    ) -> PortWatcherOptions:
        """
        *Description*
        
        Creates a new set of PortWatcherOptions with the given options.

        *Parameters*
        
        options : PortWatcherOptions
            The options to use as a base.
        plot_dir : str, optional
            The directory to save the plots in, by default None
        raw_data_dir : str, optional
            The directory to save the raw data in, by default None
        """
        # Setup
        new_plot_dir = options.plotting.base_directory
        if plot_dir is not None: 
            new_plot_dir = plot_dir

        new_raw_data_dir = options.raw_data.base_directory
        if raw_data_dir is not None: 
            new_raw_data_dir = raw_data_dir

        # Return the new options
        return PortWatcherOptions(
            plotting=PortWatcherPlottingOptions(
                plot_arrangement=options.plotting.plot_arrangement,
                plot_dpi=options.plotting.plot_dpi,
                save_to_file=options.plotting.save_to_file,
                file_format=options.plotting.file_format,
                figure_naming_convention=options.plotting.figure_naming_convention,
            ),
            raw_data=PortWatcherRawDataOptions(
                save_to_file=options.raw_data.save_to_file,
                base_directory=new_raw_data_dir,
                file_format=options.raw_data.file_format,
            ),
        )

    def __del__(self):
        """
        *Description*
        
        Destructor for the Diagram Watcher.
        Will plot the data from all of our loggers if we have access to the diagram context.
        """
        # Setup

        is_ready_to_plot = self.diagram is not None

        if not is_ready_to_plot:
            return # Return early if we don't have access to the diagram context

        # Upon deletion, we will PLOT the data from all of our loggers
        # if we have access to the diagram context
        self.save_figures()
        self.save_raw_data()

        # Close all handlers in the logger and the remove them
        for handler in self.logger.handlers:
            handler.close()
            self.logger.removeHandler(handler)

    def create_logger(self) -> logging.Logger:
        """
        *Description*

        Configures the "activity summary" a log of brom's activity.
        
        *Returns*

        logger: logging.Logger
            The configured logger for LOG MESSAGES.
            In other words, this is not a logger of signals from the diagram.
        """
        # Setup
        options = self.options

        # Create the basic logger
        logger = logging.getLogger("brom_drake.DiagramWatcher")
        for handler in logger.handlers:
            # Remove all existing handlers
            logger.removeHandler(handler)
        
        # Create a file handler, if none exists
        
        # Create a logging directory if it does not exist
        watcher_outputs_base_directory = Path(options.base_directory)
        watcher_outputs_base_directory.mkdir(parents=True, exist_ok=True)

        # Create a file handler
        file_handler = logging.FileHandler(
            filename=options.base_directory + "/activity_summary.log",
            mode='w',  # Append mode
        )
        file_handler.setLevel(logging.DEBUG)

        # Create a formatter and set it for the file handler
        formatter = logging.Formatter(
            "%(asctime)s | %(levelname)s | %(name)s | %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        file_handler.setFormatter(formatter)

        # Add the file handler to the logger
        logger.addHandler(file_handler)

        # Create a terminal handler
        terminal_handler = logging.StreamHandler()
        terminal_handler.setLevel(logging.WARNING)  # Set to WARNING to avoid cluttering terminal with INFO messages
        terminal_handler.setFormatter(formatter)
        logger.addHandler(terminal_handler)

        # Avoid duplicate logs
        logger.propagate = False

        # Make sure the logger responds to all messages of level DEBUG and above
        logger.setLevel(logging.DEBUG)  # Set to DEBUG to capture all messages

        return logger

    def check_targets(
        self,
        targets: List[DiagramTarget],
        eligible_systems: List[Union[MultibodyPlant, AffineSystem, LeafSystem]],
    ) -> List[Union[MultibodyPlant, AffineSystem, LeafSystem]]:
        """
        *Description*
        
        Finds the systems specified by the targets list that
        we want to watch/monitor.
        We will try to ignore all systems that are:
        - Scene Graphs
        - Loggers
        and raise an error if the target is not found in the eligible systems.

        *Parameters*
        
        targets : List[DiagramTarget]
            The targets that we want to monitor.
        eligible_systems : List[Union[MultibodyPlant, AffineSystem, LeafSystem]]
            The systems that are eligible for monitoring.
        
        """

        # Find all the systems that are eligible for logging
        eligible_system_dict = {
            system.get_name(): system for system in eligible_systems
        }

        # Search for each target in eligible_systems
        targeted = []
        for target in targets:
            # Check if the target name is in the eligible systems
            if target.name not in eligible_system_dict.keys():
                raise UnrecognizedTargetError(target, eligible_system_dict.keys())

            # If it is, then also check that the port index is correct
            if target.ports is None:
                continue  # No need to check things if ports is None

            num_ports_in_target = eligible_system_dict[target.name].num_output_ports()
            for port_index in target.ports:
                if port_index < 0 or port_index >= num_ports_in_target:
                    raise ValueError(f"Port index {port_index} is out of bounds for system {target.name} (only {num_ports_in_target} ports exist)")

        # All checks passed!
        pass

    def find_eligible_systems(
        self,
        builder: DiagramBuilder
    ) -> List[Union[MultibodyPlant, AffineSystem, LeafSystem]]:
        """
        *Description*

        Finds all the systems that are eligible for logging.
        We want to ignore all systems that are:

        - Scene Graphs
        - Loggers

        """

        # Find all the systems that are eligible for logging
        eligible_systems = []

        self.logger.info("Finding all eligible systems for logging...")
        for system in builder.GetSystems():
            if type(system) in INELIGIBLE_SYSTEM_TYPES:
                self.logger.warning(
                    f"System {system.get_name()} (of type {type(system)}) is not eligible for logging! Skipping..."
                )
                continue

            # Otherwise add to list
            eligible_systems.append(system)
            self.logger.info(
                f"System {system.get_name()} (of type {type(system)}) is eligible for logging with the watcher."
            )

        return eligible_systems

    def get_smart_targets(
        self,
        subject: DiagramBuilder,
        targets: List[DiagramTarget],
    ) -> List[DiagramTarget]:
        """
        *Description*

        For each target with None ports, we will try to
        "smartly" create the targets that we want to monitor.

        *Parameters*

        subject : DiagramBuilder
            The diagram builder that contains the systems.

        targets : List[DiagramTarget]
            The targets that we want to monitor.
        
        *Returns*

        List[DiagramTarget]
            The list of targets with inferred ports.
        """
        # Setup
        smart_targets = []

        for target in targets:
            if target.ports is not None:
                smart_targets.append(target)
                continue

            # If ports is None, then we will try to "smartly" create the ports
            system = subject.GetSubsystemByName(target.name)

            # TODO: Add support for investigating output ports
            num_ports = system.num_output_ports()
            if num_ports == 0:
                self.logger.warning(f"System {target.name} has no output ports! Skipping...")
                continue

            output_ports_to_watch = [port_index for port_index in range(num_ports)]

            # Add a target with the connected ports to the smart_targets list
            smart_targets.append(
                DiagramTarget(target.name, output_ports_to_watch),
            )

        # Log the list of inferred targets
        self.logger.info(f"Found {len(smart_targets)} inferred targets:")
        for idx, target in enumerate(smart_targets):
            self.logger.info(f"{idx}: {target.name} - {target.ports}")

        return smart_targets

    def save_figures(self):
        """
        *Description*

        Saves all the figures made from plotting data
        currently saved in the known port watchers.
        """
        # Announce Saving Figures has started
        self.logger.info("Saving figures...")

        # Algorithm
        for system_name in self.port_watchers:
            system_ii = self.diagram.GetSubsystemByName(system_name)
            ports_on_ii = self.port_watchers[system_name]

            self.logger.info(f"Saving figures for system {system_name}...")

            for port_name in ports_on_ii:
                temp_port_watcher = ports_on_ii[port_name]
                temp_plotting_options = temp_port_watcher.options.plotting
                
                # Plot only if the PortWatcher flag is set
                if temp_plotting_options.save_to_file: 
                    try:
                        temp_port_watcher.save_all_figures(self.diagram_context)
                        self.logger.info(f"Saved figures for port {port_name} on system {system_name}")
                    except Exception as e:
                        self.logger.error(f"Failed to save figures for port {port_name} on system {system_name}: {e}")
                else:
                    self.logger.info(f"Skipped plotting for port {port_name} on system {system_name} (flag not set)")

    def save_raw_data(self):
        """
        *Description*

        Saves all the raw data from the port watchers.
        """
        # Announce Saving Raw Data has started
        self.logger.info("Saving raw data...")

        # Algorithm
        for system_name in self.port_watchers:
            system_ii = self.diagram.GetSubsystemByName(system_name)
            ports_on_ii = self.port_watchers[system_name]

            self.logger.info(f"Saving raw data for system {system_name}...")

            for port_name in ports_on_ii:
                temp_port_watcher = ports_on_ii[port_name]
                temp_raw_data_options = temp_port_watcher.options.raw_data

                # Save raw data only if the PortWatcher flag is set
                if temp_raw_data_options.save_to_file:
                    try:
                        temp_port_watcher.save_raw_data(self.diagram_context)

                        self.logger.info(f"Saved raw data for port {port_name} on system {system_name}")

                    except Exception as e:
                        self.logger.error(f"Failed to save raw data for port {port_name} on system {system_name}: {e}")

                else:
                    self.logger.info(f"Skipped saving raw data for port {port_name} on system {system_name} (flag not set)")



from .search import (
    find_all_systems_with_input_port,
    find_all_systems_with_output_port,
)
from .leaf_systems.bool_to_vec_system import BoolToVectorSystem
from .leaf_systems.end_effector_wrench_calculator import EndEffectorWrenchCalculator
from .leaf_systems.rigid_transform_to_vector_system.system import RigidTransformToVectorSystem
from brom_drake.utils.leaf_systems.flexible_port_switch import FlexiblePortSwitch
from brom_drake.utils.leaf_systems.network_fsm.networkx_fsm import NetworkXFSM
from brom_drake.utils.leaf_systems.network_fsm.fsm_edge_definition import FSMEdgeDefinition
from brom_drake.utils.leaf_systems.network_fsm.fsm_output_definition import FSMOutputDefinition
from brom_drake.utils.leaf_systems.network_fsm.fsm_transition_condition import FSMTransitionConditionType, FSMTransitionCondition

from .constants import Performer, MotionPlan
from .collision_checking import using_point_pair_penetration
from .ground import AddGround, GroundShape
from .triads import AddMultibodyTriad, AddTriad

from .pick_and_place.phase import PickAndPlacePhase
from .pick_and_place.target_description import PickAndPlaceTargetDescription
from .puppetmaker import Puppetmaker, PuppetmakerConfiguration, PuppetSignature

__all__ = [
    "AddMultibodyTriad",
    "AddTriad",
    "BoolToVectorSystem",
    "using_point_pair_penetration",
    "EndEffectorWrenchCalculator",
    "find_all_systems_with_input_port",
    "find_all_systems_with_output_port",
    "FlexiblePortSwitch",
    "FSMEdgeDefinition",
    "FSMOutputDefinition",
    "FSMTransitionCondition",
    "FSMTransitionConditionType",
    "Performer",
    "MotionPlan",
    "NetworkXFSM",
    "PickAndPlacePhase",
    "PickAndPlaceTargetDescription",
    "Puppetmaker",
    "PuppetmakerConfiguration",
    "PuppetSignature",
    "AddGround",
    "GroundShape",
    "RigidTransformToVectorSystem",
]


"""
convenience.py
Description:

    This file contains a couple of convenience functions that are useful for the user of the URDF conversion
    feature.
"""

from typing import List, Union
from pathlib import Path

# Internal Imports
from .drake_ready_urdf_converter.converter import DrakeReadyURDFConverter
from brom_drake.file_manipulation.urdf.drake_ready_urdf_converter.config import (
    MeshReplacementStrategy, MeshReplacementStrategies, DrakeReadyURDFConverterConfig
)

def drakeify_my_urdf(
    urdf_file_path: Union[str, Path],
    overwrite_old_logs: bool = False,
    overwrite_old_models: bool = False,
    log_file_name: str = "conversion.log",
    collision_mesh_replacement_strategy: MeshReplacementStrategy = MeshReplacementStrategy.kWithObj,
    replace_colors_with: Union[None, List[float]] = None,
) -> Path:
    """
    *Description*
    
    This function provides a convenience function for the user to convert a URDF file
    into a "Drake-ready" URDF file.

    *Parameters*
    
    urdf_file_path: str
        A string representing the path to the URDF file that you would like to convert.

    overwrite_old_models: bool (optional)
        A boolean flag that indicates whether or not to overwrite old models.
        Default is False.

    overwrite_old_logs: bool (optional)
        A boolean flag that indicates whether or not to overwrite old logs.
        Default is False.

    log_file_name: str (optional)
        A string representing the name of the log file.
        Default is "conversion.log".

    collision_mesh_replacement_strategy: MeshReplacementStrategy (optional)
        An enum representing the strategy for replacing collision meshes.
        Default is MeshReplacementStrategy.kWithObj.

    *Returns*

    Path
        The path to the newly converted URDF file.
    """
    # Input Processing
    if not isinstance(urdf_file_path, str) and not isinstance(urdf_file_path, Path):
        raise ValueError(
            f"urdf_file_path must be of type str or Path; received {urdf_file_path} of type {type(urdf_file_path)}."
        )

    if isinstance(urdf_file_path, str):
        urdf_file_path = Path(urdf_file_path)

    # Create config for converter
    config = DrakeReadyURDFConverterConfig(
        overwrite_old_logs=overwrite_old_logs,
        overwrite_old_models=overwrite_old_models,
        log_file_name=log_file_name,
        mesh_replacement_strategies=MeshReplacementStrategies(
            collision_meshes=collision_mesh_replacement_strategy,
        ),
        replace_colors_with=replace_colors_with,
    )

    # Use converter
    converter = DrakeReadyURDFConverter(
        urdf_file_path,
        config=config,
    )

    # Convert the URDF
    new_urdf_path = converter.convert_urdf()

    return new_urdf_path



import numpy as np
from pydrake.all import (
    Cylinder,
    Frame,
    GeometryInstance,
    MakePhongIllustrationProperties,
    RigidTransform,
    RotationMatrix,
    SceneGraph,
)

def AddTriad(
    source_id,
    frame_id,
    scene_graph: SceneGraph,
    length: float = 0.25,
    radius: float = 0.01,
    opacity: float = 1.0,
    X_FT: RigidTransform = RigidTransform(),
    name: str= "frame",
):
    """
    Adds illustration geometry representing the coordinate frame, with the
    x-axis drawn in red, the y-axis in green and the z-axis in blue. The axes
    point in +x, +y and +z directions, respectively.

    Args:
      source_id: The source registered with SceneGraph.
      frame_id: A geometry::frame_id registered with scene_graph.
      scene_graph: The SceneGraph with which we will register the geometry.
      length: the length of each axis in meters.
      radius: the radius of each axis in meters.
      opacity: the opacity of the coordinate axes, between 0 and 1.
      X_FT: a RigidTransform from the triad frame T to the frame_id frame F
      name: the added geometry will have names name + " x-axis", etc.
    """
    # x-axis
    X_TG = RigidTransform(RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0])
    geom = GeometryInstance(
        X_FT.multiply(X_TG), Cylinder(radius, length), name + " x-axis"
    )
    geom.set_illustration_properties(
        MakePhongIllustrationProperties([1, 0, 0, opacity])
    )
    scene_graph.RegisterGeometry(source_id, frame_id, geom)

    # y-axis
    X_TG = RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0])
    geom = GeometryInstance(
        X_FT.multiply(X_TG), Cylinder(radius, length), name + " y-axis"
    )
    geom.set_illustration_properties(
        MakePhongIllustrationProperties([0, 1, 0, opacity])
    )
    scene_graph.RegisterGeometry(source_id, frame_id, geom)

    # z-axis
    X_TG = RigidTransform([0, 0, length / 2.0])
    geom = GeometryInstance(
        X_FT.multiply(X_TG), Cylinder(radius, length), name + " z-axis"
    )
    geom.set_illustration_properties(
        MakePhongIllustrationProperties([0, 0, 1, opacity])
    )
    scene_graph.RegisterGeometry(source_id, frame_id, geom)


def AddMultibodyTriad(
    frame: Frame,
    scene_graph: SceneGraph,
    length: float = 0.25,
    radius: float = 0.01,
    opacity: float = 1.0
):
    plant = frame.GetParentPlant()
    AddTriad(
        plant.get_source_id(),
        plant.GetBodyFrameIdOrThrow(frame.body().index()),
        scene_graph,
        length,
        radius,
        opacity,
        frame.GetFixedPoseInBodyFrame(),
        name=frame.name(),
    )


from dataclasses import dataclass
import numpy as np
from pydrake.all import (
    Frame
)

@dataclass
class Configuration:
    frame_on_parent: Frame
    name: str
    frame_on_child: Frame = None
    sphere_radius: float = 0.1
    sphere_color: np.ndarray = None
    sphere_mass: float = 1e-3
    create_actuators_on_init: bool = True



"""
end_effector_target.py
Description:

    An enum useful for defining the type of target we're giving to a given manipulator.
"""

from enum import Enum

class EndEffectorTarget(Enum):
    """
    *Description*
    The type of target being given for the control of a robot's end effector.

    *Usage*

    .. code-block:: python
    
        from brom_drake.control.arms.end_effector_target import EndEffectorTarget
        # or
        # from brom_drake.all import EndEffectorTarget

        # Selecting the pose target type
        ee_target_type = EndEffectorTarget.kPose
    """
    kPose = 1   # Pose (position and orientation) expressed as a 7-element vector
                # of the End Effector w.r.t. the base of the arm.
    kTwist = 2
    kWrench = 3



from .dynamic import OfflineDynamicMotionPlanningProduction
from .kinematic import KinematicMotionPlanningProduction

__all__ = ["DynamicMotionPlanningProduction", "KinematicMotionPlanningProduction"]


"""
DiagramTarget.py
Description:

    Creates a DiagramTarget class that can be used to monitor the state of a Diagram.
"""
from typing import NamedTuple, List, Union

class DiagramTarget(NamedTuple):
    """
    *Description*
    
    The target on a Diagram to record data from. 
    The object should define both:
    
    1. The LeafSystem's name, and
    2. The ports to monitor (if None, all ports are monitored).

    *Parameters*
    
    name: str
        Name of the LeafSystem to monitor

    ports: Union[None, List[int]]
        List of port indices to monitor. If None, all ports are monitored

    *Usage*

    .. code-block:: python

        from brom_drake.DiagramTarget import DiagramTarget

        # Creating a DiagramTarget to monitor all ports of a system named "my_system"
        target_all_ports = DiagramTarget(name="my_system")

        # Creating a DiagramTarget to monitor specific ports of a system named "another_system"
        target_specific_ports = DiagramTarget(name="another_system", ports=[0, 2, 4])

    """
    name: str #: Name of the LeafSystem to monitor
    ports: Union[None, List[int]] = None #: List of port indices to monitor. If None, all ports are monitored



from brom_drake.productions.debug.show_me.show_me_system import ShowMeSystem
from brom_drake.productions.debug.show_me.show_me_this_model import ShowMeThisModel

__all__ = [
    "ShowMeSystem",
    "ShowMeThisModel",
]


from typing import List, Tuple, Union
from dataclasses import dataclass
import numpy as np
from pydrake.all import (
    GeometryProperties,
    IllustrationProperties,
    RigidBodyFrame,
    RigidTransform,
    RoleAssign,
    SceneGraph,
    SceneGraphInspector,
)
from pydrake.geometry import Meshcat, MeshcatVisualizer, MeshcatVisualizerParams
from pydrake.geometry import Role as DrakeRole
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph, MultibodyPlant
from pydrake.systems.framework import DiagramBuilder, Diagram, Context

from pydrake.systems.primitives import ConstantVectorSource, VectorLogSink

# Internal Imports
from brom_drake.file_manipulation.urdf.drakeify import drakeify_my_urdf
from brom_drake.robots import find_base_link_name_in
from brom_drake.productions.types.debug import BasicGraspingDebuggingProduction
from brom_drake.productions import ProductionID
from brom_drake.productions.roles.role import Role
from brom_drake.utils import Performer, AddMultibodyTriad
from brom_drake.utils.model_instances import (
    get_name_of_first_body_in_urdf,
    find_number_of_positions_in_welded_model,
)
from brom_drake.productions.debug.show_me.show_me_system import ShowMeSystem
from .config import Configuration as ShowMeThisStaticGraspConfiguration

class ShowMeThisStaticGrasp(BasicGraspingDebuggingProduction):
    def __init__(
        self,
        path_to_object: str,
        path_to_gripper: str,
        X_ObjectGripper: RigidTransform = None,
        gripper_joint_positions: Union[List[float], np.ndarray] = None,
        config: ShowMeThisStaticGraspConfiguration = ShowMeThisStaticGraspConfiguration(),
    ):
        # Call the parent constructor
        super().__init__(
            path_to_object=path_to_object,
            path_to_gripper=path_to_gripper,
            X_ObjectGrasp=X_ObjectGripper,
            meshcat_port_number=config.base.meshcat_port_number,
            time_step=config.base.time_step,
            target_body_on_gripper=config.target_body_on_gripper,
            gripper_color=config.gripper_color,
            show_gripper_base_frame=config.show_gripper_base_frame,
            show_collision_geometries=config.show_collision_geometries,
        )

        # This will define the following fields for the production:
        # - path_to_object
        # - path_to_gripper
        # - X_ObjectTarget
        # - meshcat_port_number
        # - time_step
        # - target_body_on_gripper
        # - gripper_color
        # - show_gripper_base_frame
        # - show_collision_geometries
        # - plant

        # Create joint position array
        if gripper_joint_positions is None:
            gripper_joint_positions = [0.0] * find_number_of_positions_in_welded_model(self.path_to_gripper)
        self.gripper_joint_positions = gripper_joint_positions

        # Add Name to plantPlant and Scene Graph for easy simulation
        self.plant.set_name("DemonstrateStaticGrasp_plant")

        # Create show me system for holding object in place
        self.show_me_system = None

        self.config = config

    def add_cast_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
    ) -> Tuple[Diagram, Context]:
        super().add_cast_and_build(cast, with_watcher=True)

        # Assign the diagram context to the internal show_me_system
        self.show_me_system.mutable_plant_context = self.plant.GetMyMutableContextFromRoot(
            self.diagram_context,
        )

        return self.diagram, self.diagram_context

    def add_supporting_cast(self):
        """
        Description
        -----------
        This method will add:
        - The user's object model to the builder.
        - The user's gripper model to the builder.
        - The gripper triad to the builder.
        """
        # Setup
        plant: MultibodyPlant = self.plant

        # Add the object to the builder
        self.add_manipuland_to_plant(and_weld_to=plant.world_frame())

        # Add the gripper to the builder
        X_WorldGripper = self.find_X_WorldGripper(
            X_ObjectGripper=self.X_ObjectGripper,
            target_frame_name=self.target_body_name_on_gripper,
            desired_joint_positions=self.gripper_joint_positions,
        )
        self.add_gripper_to_plant(
            and_weld_to=plant.world_frame(),
            with_X_WorldGripper=X_WorldGripper,
        )
        self.add_gripper_controller_and_connect()

        # Connect the plant to the meshcat, if requested
        if self.meshcat_port_number is not None:
            self.connect_to_meshcat()

        # Finalize the plant
        self.plant.Finalize()

    def add_gripper_controller_and_connect(self):
        # Setup
        plant: MultibodyPlant = self.plant

        # Create a system to control the gripper's actuators
        self.show_me_system = ShowMeSystem(
            plant=plant,
            model_index=self.gripper_model_index,
            desired_joint_positions=self.gripper_joint_positions,
        )
        self.builder.AddSystem(self.show_me_system)

        # Add Command to tell the gripper to stay in a particular state
        desired_joint_positions_source = self.builder.AddSystem(
            ConstantVectorSource(np.array(self.gripper_joint_positions)),
        )
        # Connect the source to the system

        self.builder.Connect(
            desired_joint_positions_source.get_output_port(0),
            self.show_me_system.get_input_port(0),
        )

    def find_X_WorldGripper(
        self,
        X_ObjectGripper: RigidTransform,
        target_frame_name: str,
        desired_joint_positions: List[float],
    ) -> RigidTransform:
        """
        Description
        -----------
        This method will return the transform from the world frame to the gripper frame.
        """
        # Setup
        desired_joint_positions = np.array(desired_joint_positions)

        # Create shadow plant containing just the gripper
        shadow_plant = MultibodyPlant(self.time_step)
        model_idcs = Parser(plant=shadow_plant).AddModels(self.path_to_gripper)
        shadow_model_idx = model_idcs[0]
        
        # Get the first body on the gripper
        name_of_first_body_in_gripper = get_name_of_first_body_in_urdf(self.path_to_gripper)

        # Weld the base link to the world frame
        shadow_plant.WeldFrames(
            shadow_plant.world_frame(),
            shadow_plant.GetFrameByName(name_of_first_body_in_gripper),
        )

        # Finalize Plant
        shadow_plant.Finalize()

        shadow_builder = DiagramBuilder()
        shadow_builder.AddSystem(shadow_plant)
        diagram = shadow_builder.Build()
        shadow_diagram_context = diagram.CreateDefaultContext()

        # Set the joint positions
        shadow_plant.SetPositions(
            shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
            shadow_model_idx,
            desired_joint_positions,
        )

        # Get the transform of the gripper's base in the static world frame
        X_GripperBase_GripperTarget = shadow_plant.EvalBodyPoseInWorld(
            shadow_plant.GetMyContextFromRoot(shadow_diagram_context),
            shadow_plant.GetBodyByName(target_frame_name),
        )

        X_Object_GripperTarget = X_ObjectGripper
        X_WorldObject = RigidTransform.Identity() # Identity

        X_Object_GripperBase = X_Object_GripperTarget.multiply(
            X_GripperBase_GripperTarget.inverse()
        )

        return X_WorldObject.multiply(X_Object_GripperBase)

    @property
    def id(self):
        return ProductionID.kDemonstrateStaticGrasp

    @property
    def suggested_roles(self) -> List[Role]:
        return []


from pydrake.common.value import AbstractValue
from pydrake.math import RigidTransform
from pydrake.systems.framework import Context, LeafSystem

class AbstractListSelectionSystem(LeafSystem):
    """
    *Description*

    This System receives selects a single element from an AbstractValue
    that contains a list. The index of the element to select is specified
    at construction time.

    *Diagram*

    The LeafSystem's input and output ports can be illustrated with the following block: ::

                            |---------------|
            list_in ------->|               |
            (List)          | List          |
                            | Selection     | ----> element_out
                            | System        |       (Element Type)
                            |               |
                            |---------------|

    """
    def __init__(
        self,
        index: int,
        output_type: type = RigidTransform,
        output_port_name: str = "element_out"
    ):
        LeafSystem.__init__(self)

        self.index = index

        # Define Input Ports
        self.list_input_port = self.DeclareAbstractInputPort(
            "list_in",
            AbstractValue.Make([output_type()])  # Placeholder; actual type will be checked at runtime
        )

        # Define Output Port
        self.DeclareAbstractOutputPort(
            output_port_name,
            lambda: AbstractValue.Make(output_type()),  # Placeholder; actual type will be checked at runtime
            self.CalcOutputElement,
        )
        
        self.set_name("ListSelectionSystem")   
        
    def CalcOutputElement(self, context: Context, output: AbstractValue):
        """
        *Description*

        This callback function computes the selected element from the input list.
        """
        # Retrieve current input value
        input_list = self.GetInputPort("list_in").Eval(context)

        assert isinstance(input_list, list), \
            f"Expected list_in port to contain \"list\" objects, but received type {type(input_list)}"
        
        # Check index validity
        if not (0 <= self.index < len(input_list)):
            raise IndexError(
                f"Index {self.index} is out of bounds for input list of length {len(input_list)}."
            )
        
        # Select the element at the specified index
        selected_element = input_list[self.index]

        # Set the output value
        output.SetFrom(
            AbstractValue.Make(selected_element)
        )